from pathlib import Path

from m4b_util.helpers import SegmentData
from m4b_util.helpers.finders import find_chapters, find_silence
from m4b_util.helpers.finders.silence_finder import SilenceLineParser


fake_input = Path("Not-a-real-file")


def test_find_silence(silences_file_path):
    """Find silences in a file."""
    expected = [
        SegmentData(id=0, start_time=00.000, end_time=5.0, backing_file=silences_file_path,
                    file_start_time=00.00, file_end_time=5.0),
        SegmentData(id=1, start_time=05.000, end_time=10.0, backing_file=silences_file_path,
                    file_start_time=05.00, file_end_time=10.0),
        SegmentData(id=2, start_time=10.000, end_time=15.0, backing_file=silences_file_path,
                    file_start_time=10.00, file_end_time=15.0),
        SegmentData(id=3, start_time=15.000, end_time=20.011, backing_file=silences_file_path,
                    file_start_time=15.00, file_end_time=20.011),
    ]
    actual = find_silence(silences_file_path, silence_duration=0.25)
    assert actual == expected


def test_find_trimmed_silence(silences_file_path):
    """Find silences in a file."""
    expected = [
        SegmentData(id=0, start_time=00.000, end_time=02.5, backing_file=silences_file_path,
                    file_start_time=00.00, file_end_time=02.5),
        SegmentData(id=1, start_time=05.000, end_time=07.5, backing_file=silences_file_path,
                    file_start_time=05.00, file_end_time=07.5),
        SegmentData(id=2, start_time=10.000, end_time=12.5, backing_file=silences_file_path,
                    file_start_time=10.00, file_end_time=12.5),
        SegmentData(id=3, start_time=15.000, end_time=17.5, backing_file=silences_file_path,
                    file_start_time=15.00, file_end_time=17.5),
    ]
    actual = find_silence(silences_file_path, silence_duration=0.25, trim_silence=True)
    assert actual == expected


def test_silence_no_silence(variable_volume_segments_file_path):
    """Scan a file without silence."""
    expected = []
    actual = find_silence(variable_volume_segments_file_path)
    assert actual == expected


def test_silence_start_end_times(silences_file_path):
    """Scan specific parts of a file."""
    expected = [
        SegmentData(id=0, start_time=04.000, end_time=05.0, backing_file=silences_file_path,
                    file_start_time=04.00, file_end_time=05.0),
        SegmentData(id=1, start_time=05.000, end_time=10.0, backing_file=silences_file_path,
                    file_start_time=05.00, file_end_time=10.0),
        SegmentData(id=2, start_time=10.000, end_time=12.6, backing_file=silences_file_path,
                    file_start_time=10.00, file_end_time=12.6),
    ]
    actual = find_silence(silences_file_path, start_time=4.0, end_time=12.6, silence_duration=0.25)
    assert actual == expected


def test_trimmed_silence_start_end_times(silences_file_path):
    """Scan specific parts of a file."""
    expected = [
        SegmentData(id=0, start_time=05.000, end_time=07.5, backing_file=silences_file_path,
                    file_start_time=05.00, file_end_time=07.5),
        SegmentData(id=1, start_time=10.000, end_time=12.6, backing_file=silences_file_path,
                    file_start_time=10.00, file_end_time=12.6),
    ]
    actual = find_silence(silences_file_path, start_time=4.0, end_time=12.6, silence_duration=0.25, trim_silence=True)
    assert actual == expected


def test_trimmed_silence_nonsilence_ending(abrupt_ending_file_path):
    """Scan a file that ends in non-silence."""
    expected = [
        SegmentData(id=0, start_time=00.000, end_time=02.5, backing_file=abrupt_ending_file_path,
                    file_start_time=00.00, file_end_time=02.5),
        SegmentData(id=1, start_time=05.000, end_time=07.5, backing_file=abrupt_ending_file_path,
                    file_start_time=05.00, file_end_time=07.5),
        SegmentData(id=2, start_time=10.000, end_time=12.5, backing_file=abrupt_ending_file_path,
                    file_start_time=10.00, file_end_time=12.5),
        SegmentData(id=3, start_time=15.000, end_time=17.51, backing_file=abrupt_ending_file_path,
                    file_start_time=15.00, file_end_time=17.51),
    ]
    actual = find_silence(abrupt_ending_file_path, silence_duration=0.25, trim_silence=True)
    assert actual == expected


def test_silence_bad_file(tmp_path):
    """Scan a non-audio file."""
    fake_file = tmp_path / "fake.m4b"
    open(fake_file, 'a')
    expected = []
    actual = find_silence(fake_file)
    assert actual == expected


def test_find_chapters(chaptered_audio_file_path):
    """Read the chapter metadata into an object."""
    expected = [
        SegmentData(id=0, start_time=00.0, end_time=02.5, title="110Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=00.0, file_end_time=02.5),
        SegmentData(id=1, start_time=02.5, end_time=05.0, title="110Hz - Soft", backing_file=chaptered_audio_file_path,
                    file_start_time=02.5, file_end_time=05.0),
        SegmentData(id=2, start_time=05.0, end_time=07.5, title="220Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=05.0, file_end_time=07.5),
        SegmentData(id=3, start_time=07.5, end_time=10.0, title="220Hz - Soft", backing_file=chaptered_audio_file_path,
                    file_start_time=07.5, file_end_time=10.0),
        SegmentData(id=4, start_time=10.0, end_time=12.5, title="330Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=10.0, file_end_time=12.5),
        SegmentData(id=5, start_time=12.5, end_time=15.0, title="330Hz - Soft", backing_file=chaptered_audio_file_path,
                    file_start_time=12.5, file_end_time=15.0),
        SegmentData(id=6, start_time=15.0, end_time=17.5, title="440Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=15.0, file_end_time=17.5),
        SegmentData(id=7, start_time=17.5, end_time=19.999, title="440Hz - Soft",
                    file_start_time=17.5, file_end_time=19.999,
                    backing_file=chaptered_audio_file_path),
    ]
    assert find_chapters(chaptered_audio_file_path) == expected


def test_chapters_start_end_times(chaptered_audio_file_path):
    """Filter out chapters based on a custom start and end time."""
    expected = [
        SegmentData(id=2, start_time=05.0, end_time=07.5, title="220Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=05.0, file_end_time=07.5),
        SegmentData(id=3, start_time=07.5, end_time=10.0, title="220Hz - Soft", backing_file=chaptered_audio_file_path,
                    file_start_time=07.5, file_end_time=10.0),
        SegmentData(id=4, start_time=10.0, end_time=12.5, title="330Hz - Loud", backing_file=chaptered_audio_file_path,
                    file_start_time=10.0, file_end_time=12.5),
    ]
    assert find_chapters(
        chaptered_audio_file_path,
        start_time=2.75,
        end_time=14.0
    ) == expected


def test_chapters_no_chapters(mp3_path):
    """Read an audio file that has no chapter data."""
    assert find_chapters(mp3_path / "1 - 110Hz.mp3") == []


def test_chapters_unreadable_file(fake_file):
    """Find no chapters in a non-audio file."""
    assert find_chapters(fake_file) == []


def test_invalid_silence_lines():
    """Make sure we won't use times that are outside the duration of the file."""
    lines = [
        "size=N/A time=00:00:02.00 bitrate=N/A",
        " silence_end: 9.5 ",
    ]
    parser = SilenceLineParser(0.0, 10.0)
    parser.parse_lines(lines)
    times = parser.get_segments()
    assert times == []


def test_trim_silence():
    """Make sure we can trim silence from the start and end of a file."""
    lines = [
        " silence_start: 1.5 ",
        " silence_end: 2.0 ",
        " silence_start: 3.5 ",
        " silence_end: 4.0 ",
        " silence_start: 5.5 ",
        " silence_end: 6.0 ",
        " silence_start: 7.5 ",
        " silence_end: 8.0 ",
        "size=N/A time=00:00:12.00 bitrate=N/A",
    ]
    parser = SilenceLineParser(0.0, 10.0, trim_silence=True)
    parser.parse_lines(lines)
    times = parser.get_segments()
    assert times == [(0.0, 1.5), (2.0, 3.5), (4.0, 5.5), (6.0, 7.5), (8.0, 10.0)]


def test_preserving_silence():
    """Make sure we can preserve silence at the end of a file."""
    lines = [
        " silence_start: 1.5 ",
        " silence_end: 2.0 ",
        " silence_start: 3.5 ",
        " silence_end: 4.0 ",
        " silence_start: 5.5 ",
        " silence_end: 6.0 ",
        " silence_start: 7.5 ",
        " silence_end: 8.0 ",
        "size=N/A time=00:00:12.00 bitrate=N/A",
    ]
    parser = SilenceLineParser(0.0, 10.0, trim_silence=False)
    parser.parse_lines(lines)
    times = parser.get_segments()
    assert times == [(0.0, 2.0), (2.0, 4.0), (4.0, 6.0), (6.0, 8.0), (8.0, 10.0)]


def test_avoiding_double_endings():
    """When we see a silence end while preserving silence, avoid adding an end marker if it is already there."""
    lines = [
        " silence_end: 2.0 ",
        " silence_end: 4.0 ",
        "size=N/A time=00:00:12.00 bitrate=N/A",
    ]
    parser = SilenceLineParser(0.0, 10.0, trim_silence=False)
    parser.parse_line(lines[0])
    parser.segment_ends.append(2.0)  # Introduce a fake end marker
    parser.parse_line(lines[1])
    times = parser.get_segments()
    assert times == [(0.0, 2.0), (2.0, 2.0), (4.0, 10.0)]



from contextlib import contextmanager
from pathlib import Path

import pytest

from m4b_util.helpers import ffprobe


def check_output_folder(output_path, expected_files=None, check_func=None):
    """Check that expected output exists.

    :param output_path: Path to the directory to check.
    :param expected_files: Files we expect to find in output_path. If unspecified, defaults to ["segment_0000.mp3"].
    :param check_func: Function for checking each file in expected_files. Defaults to asserting ffprobe can read them.

    """
    # Default check function, in case one isn't passed in.
    def default_check(input_file_path):
        """Ensure ffprobe can read the file."""
        # Add in filename to make any failure more legible
        assert input_file_path.name and ffprobe.run_probe(input_file_path)

    # Set defaults
    if expected_files is None:
        expected_files = ["segment_0000.mp3"]
    if not check_func:
        check_func = default_check

    # Check the output
    for file_name in expected_files:
        file_path = output_path / file_name
        check_func(file_path)

    # Make sure there aren't any extra files in the directory
    for file in output_path.glob("*"):
        assert (file.name in expected_files)


def assert_file_path_is_file(input_file_path):
    """Ensure each file exists and is a file."""
    assert input_file_path.is_file()


def soften_backing_file(segment_list):
    """Remove path portion of backing_file from segments, as it will change with each test."""
    for segment in segment_list:
        segment.backing_file = Path(segment.backing_file.stem)


@contextmanager
def expect_exit(expected_code=1):
    """Ensure a function exits, and writes an expected string to stdout or stderr."""
    with pytest.raises(SystemExit) as e:
        yield
    assert e.value.code == expected_code


@contextmanager
def expect_exit_with_output(capsys, expected_text, expected_code=1):
    """Ensure a function exits, and writes an expected string to stdout or stderr."""
    with expect_exit(expected_code):
        yield
    output = capsys.readouterr()
    assert (expected_text in output.out) or (expected_text in output.err)



import argparse
from pathlib import Path
import shutil
import sys
from tempfile import TemporaryDirectory


from m4b_util.helpers import cover_utils


def _parse_args():
    parser = argparse.ArgumentParser(
        prog="m4b-util cover",
        description='Manipulate Covers.'
    )
    parser.add_argument('input_file', type=str, help='Input filename')

    parser.add_argument('-e', "--extract-cover", type=str,
                        help='Extract existing cover from input. Must end in .png or .jpg')
    parser.add_argument('-a', "--apply-cover", type=str, help='Apply specified cover to input file.')
    args = parser.parse_args(sys.argv[2:])
    if not (args.extract_cover or args.apply_cover):
        parser.error("At least one task must be specified (--extract-cover or --apply-cover).")
    return args


def run():
    """Do stuff with a cover."""
    args = _parse_args()
    if args.extract_cover:
        cover_utils.extract_cover(Path(args.input_file), Path(args.extract_cover))
    if args.apply_cover:
        with TemporaryDirectory() as tmp_path:
            covered_book = Path(tmp_path) / "covered.m4b"
            cover_utils.add_cover(Path(args.input_file), Path(args.apply_cover), covered_book)
            shutil.move(covered_book, args.input_file)



"""Fixtures for expected data."""
from pathlib import Path

import pytest

from m4b_util.helpers import SegmentData


@pytest.fixture(scope='session')
def expected_data():
    """Dictionary of data expected after processing the tone files.

    For some reason, the mp3 timestamps are different from wav and m4a.
    """
    return {
        "default": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=5000\ntitle=1\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5000\nEND=10000\ntitle=2\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10000\nEND=15000\ntitle=3\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15000\nEND=20000\ntitle=4\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20000\nEND=25000\ntitle=5\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25000\nEND=30000\ntitle=6\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30000\nEND=35000\ntitle=7\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35000\nEND=40000\ntitle=8\n'
                         ),
            "chapters": [
                SegmentData(start_time=00., end_time=05., id=1, title="1", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=05., end_time=10., id=2, title="2", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=10., end_time=15., id=3, title="3", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=15., end_time=20., id=4, title="4", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=20., end_time=25., id=5, title="5", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=25., end_time=30., id=6, title="6", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=30., end_time=35., id=7, title="7", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=35., end_time=40., id=8, title="8", backing_file=Path("8 - 880Hz"))
            ],
            "chapters_doubled": [
                SegmentData(start_time=00., end_time=05., id=1, title="1", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=05., end_time=10., id=2, title="2", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=10., end_time=15., id=3, title="3", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=15., end_time=20., id=4, title="4", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=20., end_time=25., id=5, title="5", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=25., end_time=30., id=6, title="6", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=30., end_time=35., id=7, title="7", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=35., end_time=40., id=8, title="8", backing_file=Path("8 - 880Hz")),
                SegmentData(start_time=40., end_time=45., id=9, title="9", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=45., end_time=50., id=10, title="10", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=50., end_time=55., id=11, title="11", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=55., end_time=60., id=12, title="12", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=60., end_time=65., id=13, title="13", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=65., end_time=70., id=14, title="14", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=70., end_time=75., id=15, title="15", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=75., end_time=80., id=16, title="16", backing_file=Path("8 - 880Hz"))
            ]
        },
        "mp3": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=5039\ntitle=1\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5040\nEND=10079\ntitle=2\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10080\nEND=15119\ntitle=3\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15120\nEND=20159\ntitle=4\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20160\nEND=25199\ntitle=5\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25200\nEND=30239\ntitle=6\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30240\nEND=35279\ntitle=7\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35280\nEND=40319\ntitle=8\n'
                         ),
            "chapters": [
                SegmentData(start_time=00.00, end_time=05.04, id=1, title="1", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=05.04, end_time=10.08, id=2, title="2", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=10.08, end_time=15.12, id=3, title="3", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=15.12, end_time=20.16, id=4, title="4", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=20.16, end_time=25.20, id=5, title="5", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=25.20, end_time=30.24, id=6, title="6", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=30.24, end_time=35.28, id=7, title="7", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=35.28, end_time=40.32, id=8, title="8", backing_file=Path("8 - 880Hz"))
            ],
            "chapters_doubled": [
                SegmentData(start_time=00.00, end_time=05.04, id=1, title="1", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=05.04, end_time=10.08, id=2, title="2", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=10.08, end_time=15.12, id=3, title="3", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=15.12, end_time=20.16, id=4, title="4", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=20.16, end_time=25.20, id=5, title="5", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=25.20, end_time=30.24, id=6, title="6", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=30.24, end_time=35.28, id=7, title="7", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=35.28, end_time=40.32, id=8, title="8", backing_file=Path("8 - 880Hz")),
                SegmentData(start_time=40.32, end_time=45.36, id=9, title="9", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=45.36, end_time=50.40, id=10, title="10", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=50.40, end_time=55.44, id=11, title="11", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=55.44, end_time=60.48, id=12, title="12", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=60.48, end_time=65.52, id=13, title="13", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=65.52, end_time=70.56, id=14, title="14", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=70.56, end_time=75.60, id=15, title="15", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=75.60, end_time=80.64, id=16, title="16", backing_file=Path("8 - 880Hz"))
            ]
        },
        "use_filename": {
            "metadata": (';FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n'
                         'title=None\n'
                         'artist=None\n'
                         'album=None\n'
                         'date=None\n'
                         'genre=Audiobook\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=5039\ntitle=1 - 110Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=5040\nEND=10079\ntitle=2 - 220Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=10080\nEND=15119\ntitle=3 - 330Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=15120\nEND=20159\ntitle=4 - 440Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=20160\nEND=25199\ntitle=5 - 550Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=25200\nEND=30239\ntitle=6 - 660Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=30240\nEND=35279\ntitle=7 - 770Hz\n'
                         '[CHAPTER]\nTIMEBASE=1/1000\nSTART=35280\nEND=40319\ntitle=8 - 880Hz\n'
                         ),
            "chapters": [
                SegmentData(start_time=00.00, end_time=05.04, id=1, title="1 - 110Hz", backing_file=Path("1 - 110Hz")),
                SegmentData(start_time=05.04, end_time=10.08, id=2, title="2 - 220Hz", backing_file=Path("2 - 220Hz")),
                SegmentData(start_time=10.08, end_time=15.12, id=3, title="3 - 330Hz", backing_file=Path("3 - 330Hz")),
                SegmentData(start_time=15.12, end_time=20.16, id=4, title="4 - 440Hz", backing_file=Path("4 - 440Hz")),
                SegmentData(start_time=20.16, end_time=25.20, id=5, title="5 - 550Hz", backing_file=Path("5 - 550Hz")),
                SegmentData(start_time=25.20, end_time=30.24, id=6, title="6 - 660Hz", backing_file=Path("6 - 660Hz")),
                SegmentData(start_time=30.24, end_time=35.28, id=7, title="7 - 770Hz", backing_file=Path("7 - 770Hz")),
                SegmentData(start_time=35.28, end_time=40.32, id=8, title="8 - 880Hz", backing_file=Path("8 - 880Hz"))
            ]
        },
        "chaptered": {
            "metadata": (";FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n"
                         "title=Chaptered Audio\nartist=m4b-util\nalbum=Chaptered Audio\ndate=2022\ngenre=Audiobook\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=2500\ntitle=110Hz - Loud\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=2500\nEND=5000\ntitle=110Hz - Soft\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=5000\nEND=7500\ntitle=220Hz - Loud\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=7500\nEND=10000\ntitle=220Hz - Soft\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=10000\nEND=12500\ntitle=330Hz - Loud\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=12500\nEND=15000\ntitle=330Hz - Soft\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=15000\nEND=17500\ntitle=440Hz - Loud\n"
                         "[CHAPTER]\nTIMEBASE=1/1000\nSTART=17500\nEND=19999\ntitle=440Hz - Soft\n"
                         ),
            "chapters": [
                SegmentData(start_time=0.0, end_time=2.5, id=0,
                            title='110Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=0.0, file_end_time=2.5),
                SegmentData(start_time=2.5, end_time=5.0, id=1,
                            title='110Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=2.5, file_end_time=5.0),
                SegmentData(start_time=5.0, end_time=7.5, id=2,
                            title='220Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=5.0, file_end_time=7.5),
                SegmentData(start_time=7.5, end_time=10.0, id=3,
                            title='220Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=7.5, file_end_time=10.0),
                SegmentData(start_time=10.0, end_time=12.5, id=4,
                            title='330Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=10.0, file_end_time=12.5),
                SegmentData(start_time=12.5, end_time=15.0, id=5,
                            title='330Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=12.5, file_end_time=15.0),
                SegmentData(start_time=15.0, end_time=17.5, id=6,
                            title='440Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=15.0, file_end_time=17.5),
                SegmentData(start_time=17.5, end_time=19.999, id=7,
                            title='440Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=17.5, file_end_time=19.999)
            ],
            "chapters_doubled": [
                SegmentData(start_time=0.0, end_time=2.5, id=0,
                            title='110Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=0.0, file_end_time=2.5),
                SegmentData(start_time=2.5, end_time=5.0, id=1,
                            title='110Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=2.5, file_end_time=5.0),
                SegmentData(start_time=5.0, end_time=7.5, id=2,
                            title='220Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=5.0, file_end_time=7.5),
                SegmentData(start_time=7.5, end_time=10.0, id=3,
                            title='220Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=7.5, file_end_time=10.0),
                SegmentData(start_time=10.0, end_time=12.5, id=4,
                            title='330Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=10.0, file_end_time=12.5),
                SegmentData(start_time=12.5, end_time=15.0, id=5,
                            title='330Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=12.5, file_end_time=15.0),
                SegmentData(start_time=15.0, end_time=17.5, id=6,
                            title='440Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=15.0, file_end_time=17.5),
                SegmentData(start_time=17.5, end_time=19.999, id=7,
                            title='440Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=17.5, file_end_time=19.999),
                SegmentData(start_time=19.999, end_time=22.499, id=8,
                            title='110Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=0.0, file_end_time=2.5),
                SegmentData(start_time=22.499, end_time=24.999, id=9,
                            title='110Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=2.5, file_end_time=5.0),
                SegmentData(start_time=24.999, end_time=27.499, id=10,
                            title='220Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=5.0, file_end_time=7.5),
                SegmentData(start_time=27.499, end_time=29.999, id=11,
                            title='220Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=7.5, file_end_time=10.0),
                SegmentData(start_time=29.999, end_time=32.499, id=12,
                            title='330Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=10.0, file_end_time=12.5),
                SegmentData(start_time=32.499, end_time=34.999, id=13,
                            title='330Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=12.5, file_end_time=15.0),
                SegmentData(start_time=34.999, end_time=37.499, id=14,
                            title='440Hz - Loud', backing_file=Path('chaptered_audio'),
                            file_start_time=15.0, file_end_time=17.5),
                SegmentData(start_time=37.499, end_time=39.998, id=15,
                            title='440Hz - Soft', backing_file=Path('chaptered_audio'),
                            file_start_time=17.5, file_end_time=19.999),
            ],
        }
    }



"""The Splitter Class."""
from m4b_util.helpers import cover_utils
from m4b_util.helpers.parallel_ffmpeg import ParallelFFmpeg


def split(
        input_path,
        output_dir_path,
        segment_list,
        output_pattern="segment_{i:04d}.mp3",
        padding=0.0
):
    """
    Split a file into multiple, based on segments.

    Args:
        input_path (Path): Path to the input file.
        output_dir_path (Path): Path to the directory to place the output files.
        segment_list (list[Segment]): List of segments to split the file into.
        output_pattern (str): Output filename pattern (e.g. `segment_{i:04d}.mp3`)
        padding (float): Silence to add to the end of the segments once the original has been removed.
    """
    cover_utils.extract_cover(input_path, output_dir_path / "cover.png")

    # Generate task list
    tasks = list()
    for i, segment in enumerate(segment_list):
        time = segment.end_time - segment.start_time

        output_dir_path.mkdir(exist_ok=True)
        output_path = output_dir_path / output_pattern.format(i, i=i, title=segment.title)

        # Set up our FFMPEG command
        cmd = ["ffmpeg", "-ss", str(segment.start_time), "-t", str(time), "-i", input_path]

        # Check to see if we had a cover image. If so, include it
        if (output_dir_path / "cover.png").exists():
            cmd.extend(["-i", output_dir_path / "cover.png", "-map", "1:0"])

        # Add the guaranteed mappings
        cmd.extend(["-map", "0:a", "-map_chapters", "-1", "-y"])

        # If we have a title, add it to the metadata
        if segment.title:
            cmd.extend(["-metadata", f"title={segment.title}"])

        # Add padding at the end of the segment
        if padding > 0.0:
            cmd.extend(["-af", f"apad=pad_dur={padding}"])

        # Finish the command with our output path
        cmd.append(output_path)

        name = f"Splitting segment {i}"
        if segment.title:
            name += f" - {segment.title}"
        tasks.append({
            "name": name,
            "command": cmd
        })

    # Process splits in parallel
    p = ParallelFFmpeg(f"Splitting '{input_path.name}'")
    p.process(tasks)



"""PyTest Configuration."""
from pathlib import Path
from subprocess import run

import pytest

from expected_data import expected_data  # noqa

# Let pytest know we would like them to handle asserts in our helper code
pytest.register_assert_rewrite("testhelpers")


def _generate_tone_files(dir_path, extension):
    """Generate a folder of sine wave audio files.

    Creates 8 files with 2.5 seconds of pure sine wave at the base frequency, increasing by the base frequency
    each time.

    :param dir_path: Base folder in which to place the output.
    :param extension: File extension (and thus, implicitly, the format) to use for the output files.
    """
    base_frequency = 110  # Hz
    for i in range(1, 9):
        frequency = base_frequency * i
        output_path = dir_path / f"{i} - {frequency}Hz.{extension}"
        cmd = ["ffmpeg", "-f", "lavfi", "-i", f"sine=frequency={frequency}:sample_rate=48000:duration=2.5",
               "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", output_path]
        run(cmd, capture_output=True, check=True)


@pytest.fixture(scope='session')
def test_data_path():
    """Path to the test data directory."""
    # Since __file__ returns the file name, we need to call parent to get the directory
    return Path(__file__).parent.joinpath("data").absolute()


@pytest.fixture()
def wav_path(tmp_path):
    """Path to a temporary folder with a number of short wav files."""
    wav_path = tmp_path / "wav"
    wav_path.mkdir()
    _generate_tone_files(wav_path, "wav")

    return wav_path


@pytest.fixture()
def mp3_path(tmp_path):
    """Path to a temporary folder with a number of short mp3 files."""
    mp3_path = tmp_path / "mp3"
    mp3_path.mkdir()
    _generate_tone_files(mp3_path, "mp3")

    return mp3_path


@pytest.fixture()
def mp3_file_path(tmp_path):
    """A single mp3 file. 2.5s tone, 2.5s silence."""
    output_path = tmp_path / "single_file.mp3"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", output_path]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def m4a_path(tmp_path):
    """Path to a temporary folder with a number of short m4a files."""
    m4a_path = tmp_path / "m4a"
    m4a_path.mkdir()
    _generate_tone_files(m4a_path, "m4a")

    return m4a_path


@pytest.fixture()
def m4a_file_path(tmp_path):
    """A single m4a file. 2.5s tone, 2.5s silence."""
    output_path = tmp_path / "single_file.m4a"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", output_path]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def covered_audio_file(tmp_path, test_data_path):
    """Path to an m4a file with a cover image."""
    output_path = tmp_path / "covered_audio_file.m4a"
    cover_path = test_data_path / "cover.png"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-i", cover_path, "-c:v", "png",
           "-map", "0:a", "-map", "1", "-disposition:v:0", "attached_pic",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def video_only_file(tmp_path):
    """File containing a single video stream test pattern."""
    output_path = tmp_path / "video_only.mp4"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "testsrc=duration=5.3:size=qcif:rate=10", output_path]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def chaptered_audio_file_path(tmp_path):
    """File containing multiple chapters."""
    metadata = (
        ";FFMETADATA1\nmajor_brand=M4A\nminor_version=512\ncompatible_brands=M4A isomiso2\n"
        "title=Chaptered Audio\nartist=m4b-util\nalbum=Chaptered Audio\ndate=2022\ngenre=Audiobook\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=0\nEND=2499\ntitle=110Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=2500\nEND=4999\ntitle=110Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=5000\nEND=7499\ntitle=220Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=7500\nEND=9999\ntitle=220Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=10000\nEND=12499\ntitle=330Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=12500\nEND=14999\ntitle=330Hz - Soft\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=15000\nEND=17499\ntitle=440Hz - Loud\n"
        "[CHAPTER]\nTIMEBASE=1/1000\nSTART=17500\nEND=19999\ntitle=440Hz - Soft\n"
    )
    metadata_file = tmp_path / "ffmetadata"
    with open(metadata_file, 'w') as f:
        print(metadata, file=f)
    output_path = tmp_path / "chaptered_audio.m4b"
    cmd = ["ffmpeg",
           "-i", metadata_file,
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=.5",
           "-filter_complex",
           "[1][2][3][4][5][6][7][8]concat=n=8:v=0:a=1[s0]",
           "-map", "[s0]", "-map_metadata", "0", "-map_chapters", "0",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def silences_file_path(tmp_path):
    """File containing multiple segments of silence."""
    output_path = tmp_path / "file_with_silences.m4a"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex",
           "[0]apad=pad_dur=2.5[s0],[1]apad=pad_dur=2.5[s1],[2]apad=pad_dur=2.5[s2],[3]apad=pad_dur=2.5[s3],"
           "[s0][s1][s2][s3]concat=n=4:v=0:a=1[s4]", "-map", "[s4]",
           output_path
           ]

    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def abrupt_ending_file_path(tmp_path):
    """File containing multiple segments of silence, ending in non-silence."""
    output_path = tmp_path / "file_with_abrupt_end.m4a"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5",
           "-filter_complex",
           "[0]apad=pad_dur=2.5[s0],[1]apad=pad_dur=2.5[s1],[2]apad=pad_dur=2.5[s2],"
           "[s0][s1][s2][3]concat=n=4:v=0:a=1[s4]", "-map", "[s4]",
           output_path
           ]

    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def variable_volume_segments_file_path(tmp_path):
    """File containing multiple segments of louder and softer sounds."""
    output_path = tmp_path / "variable_volume_segments.m4b"
    cmd = ["ffmpeg",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=220:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=330:sample_rate=48000:duration=2.5,volume=.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=1.5",
           "-f", "lavfi", "-i", "sine=frequency=440:sample_rate=48000:duration=2.5,volume=.5",
           "-filter_complex",
           "[0][1][2][3][4][5][6][7]concat=n=8:v=0:a=1[s0]", "-map", "[s0]",
           output_path
           ]
    run(cmd, capture_output=True, check=True)
    return output_path


@pytest.fixture()
def mkv_file_path(tmp_path):
    """An mkv file, for testing files without duration metadata."""
    # Create the file
    file = tmp_path / "no_duration.mkv"
    cmd = ["ffmpeg", "-f", "lavfi", "-i", "sine=frequency=110:sample_rate=48000:duration=2.5",
           "-filter_complex", "[0]apad=pad_dur=2.5[s0]", "-map", "[s0]", file]
    run(cmd, capture_output=True, check=True)

    return file


@pytest.fixture()
def fake_file(tmp_path):
    """A file with no content."""
    file = tmp_path / "fake.file"
    open(file, 'a').close()
    return file



"""Bind Command."""
import argparse
from pathlib import Path
import sys

from rich import print

from m4b_util.helpers import Audiobook


def _parse_args():
    """Parse all arguments."""
    parser = argparse.ArgumentParser(
        description="Take a folder of audio files and output an m4b.",
        prog="m4b-util bind"
    )
    parser.add_argument('input_folder', nargs="?", type=str, help="The folder to scan for input files.")
    parser.add_argument('-a', "--author", type=str, help="Name of the author.")
    parser.add_argument('-c', "--cover", type=str, help="Image file to use as cover")
    parser.add_argument('-f', "--files", nargs='+', type=str, action='extend',
                        help="Specific files to use for the audiobook. Overrides `input_folder`.")
    parser.add_argument('-o', "--output-dir", type=str, help="Directory to put the finished audiobook.")
    parser.add_argument('-n', "--output-name", type=str, help="Filename to use for finished audiobook. Default"
                                                              " is '[Author] - [Title].m4b'.")
    parser.add_argument('-t', "--title", type=str, help="Title of the audiobook.")
    parser.add_argument("--date", type=str, help="Date to include in metadata.")
    parser.add_argument("--decode-durations", "--decode-duration", action='store_true', default=False,
                        help="Fully decode each file to determine its duration (Slower, but more accurate).")
    parser.add_argument("--show-order", action='store_true',
                        help="Show the order the files would be read in, then exit.")
    parser.add_argument("--keep-temp-files", action='store_true', help="Skip cleanup. (Debugging)")
    parser.add_argument("--use-filename", "--use-filenames", action='store_true', default=False,
                        help="Use the filename as the chapter title instead of the title from the file's metadata.")

    return parser.parse_args(sys.argv[2:])


def check_args(input_folder, files, output_dir):
    """Check the arguments for validity."""
    # We either need an input folder or a list of files.
    if not input_folder and not files:
        print("[bold red]Error:[/] You must provide either an input folder or a list of files.")
        return -1

    # Warn the user if they specified both an input folder and a list of files.
    if input_folder and files:
        print("[bold yellow]Warning:[/] Both an input folder and specific files were specified. "
              "The input folder will be ignored.")

    # Make sure the output directory exists, if it was specified.
    if output_dir and not Path(output_dir).is_dir():
        print(f"[bold red]Error:[/] '{output_dir}' is not a directory.")
        return -1


def run():
    """Entrypoint for bind subcommand."""
    args = _parse_args()

    check_args(args.input_folder, args.files, args.output_dir)

    # Set info from args
    book = Audiobook(
        author=args.author,
        cover=args.cover,
        output_name=args.output_name,
        title=args.title,
        date=args.date,
        keep_temp_files=args.keep_temp_files,
    )

    #
    if args.files:
        filelist = [Path(file) for file in args.files]

        # Print order, if applicable
        if args.show_order:
            for i, file in enumerate(filelist):
                print(f"{i:3}:\t{file.name}")
            return 0

        # Add the files to the binder
        book.add_chapters_from_filelist(
            input_files=filelist,
            use_filenames=args.use_filename,
            decode_durations=args.decode_durations
        )

    else:  # We are using the directory scanner

        # Print order, if applicable
        if args.show_order:
            for i, file in enumerate(book.scan_dir(args.input_folder)):
                print(f"{i:3}:\t{file.name}")
            return 0

        # Add the files to the binder
        if not args.input_folder:
            print("[red]Error:[/] No input folder specified.")
            return -1
        book.add_chapters_from_directory(
            input_dir=args.input_folder,
            use_filenames=args.use_filename,
            decode_durations=args.decode_durations
        )

    # Run the binder
    output_path = Path()
    if args.output_dir:
        output_path = Path(args.output_dir)
    output_path = output_path / book.suggested_file_name
    if book.bind(output_path=output_path):
        # Tell the user where it is.
        print(f"[cyan]Writing '[yellow]{output_path}[/]'")

    return 0



"""Chapter Finder."""
import re

from m4b_util.helpers import ffprogress
from m4b_util.helpers.segment_data import SegmentData

FAKE_DURATION = 10000000000000.0  # Default to a very large number until we know the real duration.
silence_start_re = re.compile(r' silence_start: (?P<start>[0-9]+(\.?[0-9]*))')
silence_end_re = re.compile(r' silence_end: (?P<end>[0-9]+(\.?[0-9]*))')
total_duration_re = re.compile(
    r'size=[^ ]+ time=(?P<hours>[0-9]{2}):(?P<minutes>[0-9]{2}):(?P<seconds>[0-9\.]{5}) bitrate=')


class SilenceLineParser:
    """Utility class for parsing ffmpeg's silencedetect filter output."""
    def __init__(self, start_time, end_time, trim_silence=False):
        self.start_time = start_time or 0.0
        self.end_time = end_time or FAKE_DURATION
        self.trim_silence = trim_silence
        self.segment_starts = []
        self.segment_ends = []
        self.duration = FAKE_DURATION

    def parse_line(self, line):
        """Parse a single line of silencedetect output."""
        self._check_for_start(line)
        self._check_for_end(line)
        self._check_for_duration(line)

    def _check_for_start(self, line):
        """Check for silence start. Only applies if we're trimming silence."""
        if self.trim_silence:
            silence_start_match = silence_start_re.search(line)
            if silence_start_match:
                timestamp = self.start_time + float(silence_start_match.group('start'))
                if self.start_time < timestamp < self.end_time:
                    self.segment_ends.append(timestamp)
                    if len(self.segment_starts) == 0:
                        # Started with non-silence.
                        self.segment_starts.append(self.start_time)

    def _check_for_duration(self, line):
        total_duration_match = total_duration_re.search(line)
        if total_duration_match:
            hours = int(total_duration_match.group('hours'))
            minutes = int(total_duration_match.group('minutes'))
            seconds = float(total_duration_match.group('seconds'))
            self.duration = hours * 3600 + minutes * 60 + seconds + self.start_time
            if self.end_time == FAKE_DURATION:
                self.end_time = self.duration

    def _check_for_end(self, line):
        silence_end_match = silence_end_re.search(line)
        if silence_end_match:
            timestamp = self.start_time + float(silence_end_match.group('end'))
            # end_time could theoretically be set further than the actual duration, so we need to check both.
            if timestamp < self.end_time and timestamp < self.duration:
                if not self.trim_silence:
                    # If we're not trimming silence, segments won't get marked as ended elsewhere.
                    if len(self.segment_starts) == 0:
                        self.segment_starts.append(self.start_time)
                    # If we have an existing segment, it will need to be marked as ending here.
                    if len(self.segment_starts) > len(self.segment_ends):
                        self.segment_ends.append(timestamp)

                # Mark the new segment too
                self.segment_starts.append(timestamp)

    def parse_lines(self, lines):
        """Parse a list of lines."""
        for line in lines:
            self.parse_line(line)

    def get_segments(self):
        """Return the list of segments."""
        if len(self.segment_starts) > len(self.segment_ends):
            # If we've got a valid segment start with no segment end, we can assume the input ended without silence.
            # That means we need to add the global end time as the final segment end time.
            if self.segment_starts[-1] < self.duration:
                self.segment_ends.append(self.end_time)
            else:
                # If we do have a segment start that is beyond the duration, then we should just remove it.
                self.segment_starts.pop()

        segment_list = list(zip(self.segment_starts, self.segment_ends))
        return segment_list


def find_silence(
        input_path,
        start_time=None,
        end_time=None,
        silence_duration=3.0,
        silence_threshold=-35,
        trim_silence=False):
    """
    Finds silence in a file and generates a list of SegmentData's representing the non-silence portions.

    Args:
        input_path (Path): Path to the input file.
        start_time (float): Time to start processing from.
        end_time (float): Time to stop processing at.
        silence_duration (float): Duration of silence to detect.
        silence_threshold (float): Silence threshold in dB.
        trim_silence (bool): If True, silence will be trimmed from the start and end of the segments.
    """
    # Run ffmpeg's silencedetect filter
    cmd = ["ffmpeg"]
    if start_time:
        cmd.extend(["-ss", str(start_time)])
    cmd.extend(["-i", input_path])
    if end_time:
        end_time_dur = end_time - start_time
        cmd.extend(["-t", str(end_time_dur)])
    cmd.extend([
        "-filter_complex",
        f"[0]silencedetect=d={silence_duration}:n={silence_threshold}dB[s0]",
        "-map", "[s0]",
        "-f", "null",
        "-"
    ])
    ff = ffprogress.run(cmd, "Detecting Silence")

    # Check to see if ffmpeg exited abnormally
    lines = []
    if ff:
        lines = ff.output.splitlines()

    # Parse the output
    line_parser = SilenceLineParser(start_time, end_time, trim_silence)
    line_parser.parse_lines(lines)
    times = line_parser.get_segments()

    # Generate SegmentData list
    retval = list()
    for i, (segment_start, segment_end) in enumerate(times):
        retval.append(SegmentData(
            start_time=segment_start,
            end_time=segment_end,
            id=i,
            backing_file=input_path,
            file_start_time=segment_start,
            file_end_time=segment_end
        ))
    return retval



"""Chapter Finder."""
from m4b_util.helpers import ffprobe
from m4b_util.helpers.segment_data import SegmentData


def find_chapters(input_path, start_time=None, end_time=None):
    """Read chapter metadata from a file and generates a list of matching SegmentData's."""
    # Process defaults here, since we often get passed argparse values directly.
    if start_time is None:
        start_time = 0.0
    if end_time is None:
        end_time = 100000000000000.0
    probe = ffprobe.run_probe(input_path)
    if probe is None:
        return []  # If we can't read the file, then we didn't find any chapters.
    chapter_list = list()
    for chapter in probe.chapters:
        title = chapter.get("tags", dict()).get("title")
        if float(chapter['start_time']) >= start_time and float(chapter['end_time']) <= end_time:
            chapter_list.append(SegmentData(
                start_time=float(chapter['start_time']),
                end_time=float(chapter['end_time']),
                id=chapter['id'],
                title=title,
                backing_file=input_path,
                file_start_time=float(chapter['start_time']),
                file_end_time=float(chapter['end_time'])
            ))
    return chapter_list



"""Cover Subcommand Tests."""
import filecmp
import shutil
from unittest import mock

import testhelpers

from m4b_util.helpers import ffprobe
from m4b_util.subcommands import cover


def _run_cover_cmd(arg_list):
    """Patch the runtime arguments, then run the split command."""
    argv_patch = ["m4b-util", "cover"]
    argv_patch.extend(arg_list)

    with mock.patch("sys.argv", argv_patch):
        cover.run()


def test_extract(tmp_path, test_data_path, covered_audio_file):
    """Extract a cover."""
    cover_path = tmp_path / "out.png"
    _run_cover_cmd([str(covered_audio_file), "-e", str(cover_path)])
    assert cover_path.is_file()
    assert filecmp.cmp(test_data_path / "cover.png", cover_path, shallow=False)


def test_apply(test_data_path, m4a_file_path):
    """Add a new cover."""
    cover_path = test_data_path / "cover.png"
    _run_cover_cmd([str(m4a_file_path), "-a", str(cover_path)])
    probe = ffprobe.run_probe(m4a_file_path)
    assert probe
    cover_streams = [stream for stream in probe.data.get('streams', list()) if stream.get('codec_name') == 'png']
    assert len(cover_streams) == 1
    assert cover_streams[0].get('disposition', {}).get("attached_pic") == 1


def test_overwrite(tmp_path, test_data_path, covered_audio_file):
    """Overwrite an exist cover."""
    new_cover_path = test_data_path / "cover2.png"
    original_file = tmp_path / "original.m4a"
    shutil.copy(covered_audio_file, original_file)
    _run_cover_cmd([str(covered_audio_file), "-a", str(new_cover_path)])
    probe = ffprobe.run_probe(covered_audio_file)
    assert probe
    cover_streams = [stream for stream in probe.data.get('streams', list()) if stream.get('codec_name') == 'png']
    assert len(cover_streams) == 1
    assert cover_streams[0].get('disposition', {}).get("attached_pic") == 1
    assert not filecmp.cmp(original_file, covered_audio_file, shallow=False)


def test_no_task_given(capsys):
    """Alert the user if no tasks are specified."""
    with testhelpers.expect_exit_with_output(
            capsys,
            expected_code=2,
            expected_text="At least one task must be specified"
    ):
        _run_cover_cmd(["Not-a-real.file"])



# Standard Library

# Third Party
from lark import Transformer

# Local


class FFMetadataTransformer(Transformer):
    """Transforms a Lark parse tree into a dictionary of metadata items and chapters."""
    def __init__(self):
        self.chapters = list()
        self.global_data = dict()

    def metadata(self, data):
        """Split the global metadata from the chapters."""
        for item in data:
            if isinstance(item, dict):
                self.chapters.append(item)
            else:
                self.global_data[item[0]] = item[1]
        return {"global": self.global_data, "chapters": self.chapters}
    chapter = dict
    data = tuple
    KEY = str
    VALUE = str


FFMETADATA_GRAMMAR = r"""
?metadata: _WS* _HEADER _WS* data* chapter*
chapter: _CHAPTER_MARKER _WS* data*
data: KEY _EQUAL_SIGN VALUE _WS*
KEY: /[^\s=]+/
VALUE: /[^\n]+/

_HEADER: ";FFMETADATA1"i
_CHAPTER_MARKER: "[CHAPTER]"i
_EQUAL_SIGN: "="

_NEWLINE: NEWLINE
_WS: WS
%import common.NEWLINE
%import common.WS
"""

FFMETADATA_TERMINATOR_FRIENDLY_NAMES = {
    "_HEADER": "Header (ie. ';FFMETADATA1')",
    "_CHAPTER_MARKER": "Chapter Marker (ie. '[CHAPTER]')",
    "_EQUAL_SIGN": "Equal Sign (ie. '=')",
    "KEY": "Metadata item name",
    "VALUE": "Metadata item value",
    "_NEWLINE": "<New Line>",
    "_WS": "<Whitespace>",
}



"""Slide command tests."""
import shutil
from unittest import mock

import testhelpers

from m4b_util.helpers import SegmentData
from m4b_util.helpers.finders import find_chapters
from m4b_util.subcommands import slide


def _slide_and_check_segments(arg_list, input_path, expected_segments):
    """Run the split command, find chapters in the output, and check against what we expect."""
    _run_slide_cmd(arg_list)
    segment_list = find_chapters(input_path)
    # Remove backing file info. We aren't testing that.
    for segment in segment_list:
        segment.backing_file = None
    assert segment_list == expected_segments


def _run_slide_cmd(arg_list):
    """Patch the runtime arguments, then run the split command."""
    argv_patch = ["m4b-util", "slide"]
    argv_patch.extend(arg_list)

    with mock.patch("sys.argv", argv_patch):
        slide.run()


def test_slide_forward(tmp_path, chaptered_audio_file_path):
    """Slide all chapters forward."""
    cmd = [str(chaptered_audio_file_path), "-d", "1.0"]
    expected = [
        SegmentData(id=0, title="110Hz - Loud",
                    start_time=00.0, file_start_time=00.0,
                    end_time=03.5, file_end_time=03.5),
        SegmentData(id=1, title="110Hz - Soft",
                    start_time=03.5, file_start_time=03.5,
                    end_time=06.0, file_end_time=06.0),
        SegmentData(id=2, title="220Hz - Loud",
                    start_time=06.0, file_start_time=06.0,
                    end_time=08.5, file_end_time=08.5),
        SegmentData(id=3, title="220Hz - Soft",
                    start_time=08.5, file_start_time=08.5,
                    end_time=11.0, file_end_time=11.0),
        SegmentData(id=4, title="330Hz - Loud",
                    start_time=11.0, file_start_time=11.0,
                    end_time=13.5, file_end_time=13.5),
        SegmentData(id=5, title="330Hz - Soft",
                    start_time=13.5, file_start_time=13.5,
                    end_time=16.0, file_end_time=16.0),
        SegmentData(id=6, title="440Hz - Loud",
                    start_time=16.0, file_start_time=16.0,
                    end_time=18.5, file_end_time=18.5),
        SegmentData(id=7, title="440Hz - Soft",
                    start_time=18.5, file_start_time=18.5,
                    end_time=19.999, file_end_time=19.999),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)


def test_slide_backward(tmp_path, chaptered_audio_file_path):
    """Slide all chapters backwards, preserving start and end times."""
    cmd = [str(chaptered_audio_file_path), "-d", "-1.0"]
    expected = [
        SegmentData(id=0, title="110Hz - Loud",
                    start_time=00.0, file_start_time=00.0,
                    end_time=01.5, file_end_time=01.5),
        SegmentData(id=1, title="110Hz - Soft",
                    start_time=01.5, file_start_time=01.5,
                    end_time=04.0, file_end_time=04.0),
        SegmentData(id=2, title="220Hz - Loud",
                    start_time=04.0, file_start_time=04.0,
                    end_time=06.5, file_end_time=06.5),
        SegmentData(id=3, title="220Hz - Soft",
                    start_time=06.5, file_start_time=06.5,
                    end_time=09.0, file_end_time=09.0),
        SegmentData(id=4, title="330Hz - Loud",
                    start_time=09.0, file_start_time=09.0,
                    end_time=11.5, file_end_time=11.5),
        SegmentData(id=5, title="330Hz - Soft",
                    start_time=11.5, file_start_time=11.5,
                    end_time=14.0, file_end_time=14.0),
        SegmentData(id=6, title="440Hz - Loud",
                    start_time=14.0, file_start_time=14.0,
                    end_time=16.5, file_end_time=16.5),
        SegmentData(id=7, title="440Hz - Soft",
                    start_time=16.5, file_start_time=16.5,
                    end_time=19.999, file_end_time=19.999),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)


def test_slide_segment_list_forward():
    """Shift all entries in a segment list forward, preserving start and end times."""
    initial = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=05.0, file_end_time=05.0),
        SegmentData(id=1,
                    start_time=05.0, file_start_time=05.0,
                    end_time=10.0, file_end_time=10.0),
        SegmentData(id=2,
                    start_time=10.0, file_start_time=10.0,
                    end_time=15.0, file_end_time=15.0),
        SegmentData(id=3,
                    start_time=15.0, file_start_time=15.0,
                    end_time=20.0, file_end_time=20.0),
    ]
    expected = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=05.5, file_end_time=05.5),
        SegmentData(id=1,
                    start_time=05.5, file_start_time=05.5,
                    end_time=10.5, file_end_time=10.5),
        SegmentData(id=2,
                    start_time=10.5, file_start_time=10.5,
                    end_time=15.5, file_end_time=15.5),
        SegmentData(id=3,
                    start_time=15.5, file_start_time=15.5,
                    end_time=20.0, file_end_time=20.0),
    ]
    actual = slide._slide_segments(initial, 0.5)
    assert actual == expected


def test_overshoot_backing_only():
    """Overshoot backing file times."""
    initial = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=05.0, file_end_time=05.0),
        SegmentData(id=1,
                    start_time=05.0, file_start_time=05.0,
                    end_time=10.0, file_end_time=10.0),
        SegmentData(id=2,
                    start_time=10.0, file_start_time=10.0,
                    end_time=15.0, file_end_time=15.0),
        SegmentData(id=3,
                    start_time=15.0, file_start_time=15.0,
                    end_time=20.0, file_end_time=17.5),
    ]
    expected = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=07.6, file_end_time=07.6),
        SegmentData(id=1,
                    start_time=07.6, file_start_time=07.6,
                    end_time=12.6, file_end_time=12.6),
        SegmentData(id=2,
                    start_time=12.6, file_start_time=12.6,
                    end_time=20.0, file_end_time=17.5),
    ]
    actual = slide._slide_segments(initial, 2.6)
    assert actual == expected


def test_slide_segment_list_backward():
    """Shift all entries in a segment list backwards, except for the initial start time."""
    initial = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=05.0, file_end_time=05.0),
        SegmentData(id=1,
                    start_time=05.0, file_start_time=05.0,
                    end_time=10.0, file_end_time=10.0),
        SegmentData(id=2,
                    start_time=10.0, file_start_time=10.0,
                    end_time=15.0, file_end_time=15.0),
        SegmentData(id=3,
                    start_time=15.0, file_start_time=15.0,
                    end_time=20.0, file_end_time=20.0),
    ]
    expected = [
        SegmentData(id=0,
                    start_time=00.0, file_start_time=00.0,
                    end_time=04.5, file_end_time=04.5),
        SegmentData(id=1,
                    start_time=04.5, file_start_time=04.5,
                    end_time=09.5, file_end_time=09.5),
        SegmentData(id=2,
                    start_time=09.5, file_start_time=09.5,
                    end_time=14.5, file_end_time=14.5),
        SegmentData(id=3,
                    start_time=14.5, file_start_time=14.5,
                    end_time=20.0, file_end_time=20.0),
    ]
    actual = slide._slide_segments(initial, -0.5)
    assert actual == expected


def test_slide_segments_no_backing():
    """Shift all entries in a segment list forward, except for the final end time."""
    initial = [
        SegmentData(id=0, start_time=00.0, end_time=05.0),
        SegmentData(id=1, start_time=05.0, end_time=10.0),
        SegmentData(id=2, start_time=10.0, end_time=15.0),
        SegmentData(id=3, start_time=15.0, end_time=20.0),
    ]
    expected = [
        SegmentData(id=0, start_time=00.0, end_time=05.5),
        SegmentData(id=1, start_time=05.5, end_time=10.5),
        SegmentData(id=2, start_time=10.5, end_time=15.5),
        SegmentData(id=3, start_time=15.5, end_time=20.0),
    ]
    actual = slide._slide_segments(initial, 0.5)
    assert actual == expected


def test_zero_segments():
    """Return immediately if presented with a zero segment list."""
    initial = []
    expected = []
    actual = slide._slide_segments(initial, 1)
    assert initial == expected == actual


def test_zero_slide_duration():
    """Return immediately if presented with a zero slide durationn."""
    initial = ["Not-a-val"]
    expected = ["Not-a-val"]
    actual = slide._slide_segments(initial, 0)
    assert initial == expected == actual


def test_no_chapters(tmp_path, silences_file_path, capsys):
    """Alert the user if there are no chapters to slide."""
    original_file = tmp_path / "original.m4b"
    shutil.copy(silences_file_path, original_file)
    with testhelpers.expect_exit_with_output(capsys, "No chapters found"):
        _run_slide_cmd([str(silences_file_path), "-d", "-1.0"])


def test_trim_start(tmp_path, chaptered_audio_file_path):
    """Trim a length from the beginning of a file."""
    cmd = [str(chaptered_audio_file_path), "--trim-start", "3.6"]
    expected = [
        SegmentData(id=0, title="110Hz - Soft",
                    start_time=00.0, file_start_time=00.0,
                    end_time=01.4, file_end_time=01.4),
        SegmentData(id=1, title="220Hz - Loud",
                    start_time=01.4, file_start_time=01.4,
                    end_time=03.9, file_end_time=03.9),
        SegmentData(id=2, title="220Hz - Soft",
                    start_time=03.9, file_start_time=03.9,
                    end_time=06.4, file_end_time=06.4),
        SegmentData(id=3, title="330Hz - Loud",
                    start_time=06.4, file_start_time=06.4,
                    end_time=08.9, file_end_time=08.9),
        SegmentData(id=4, title="330Hz - Soft",
                    start_time=08.9, file_start_time=08.9,
                    end_time=11.4, file_end_time=11.4),
        SegmentData(id=5, title="440Hz - Loud",
                    start_time=11.4, file_start_time=11.4,
                    end_time=13.9, file_end_time=13.9),
        SegmentData(id=6, title="440Hz - Soft",
                    start_time=13.9, file_start_time=13.9,
                    end_time=16.399, file_end_time=16.399),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)


def test_trim_too_much(tmp_path, chaptered_audio_file_path, capsys):
    """Warn the user if they try to trim more off the start than is in the file."""
    original_file = tmp_path / "original.m4b"
    shutil.copy(chaptered_audio_file_path, original_file)
    with testhelpers.expect_exit_with_output(capsys, "No chapters were left after trim."):
        _run_slide_cmd([str(chaptered_audio_file_path), "--trim-start", "50.0"])


def test_slide_too_much(tmp_path, chaptered_audio_file_path, capsys):
    """Warn the user if they try to trim more off the start than is in the file."""
    original_file = tmp_path / "original.m4b"
    shutil.copy(chaptered_audio_file_path, original_file)
    with testhelpers.expect_exit_with_output(capsys, "No chapters were left after slide."):
        _run_slide_cmd([str(chaptered_audio_file_path), "-d", "50.0"])


def test_trim_and_slide_backwards(tmp_path, chaptered_audio_file_path):
    """Trim a length from the beginning of a file while shifting everything backwards."""
    cmd = [str(chaptered_audio_file_path), "--trim-start", "3.6", "-d", "-1"]
    expected = [
        SegmentData(id=0, title="110Hz - Soft",
                    start_time=00.0, file_start_time=00.0,
                    end_time=00.4, file_end_time=00.4),
        SegmentData(id=1, title="220Hz - Loud",
                    start_time=00.4, file_start_time=00.4,
                    end_time=02.9, file_end_time=02.9),
        SegmentData(id=2, title="220Hz - Soft",
                    start_time=02.9, file_start_time=02.9,
                    end_time=05.4, file_end_time=05.4),
        SegmentData(id=3, title="330Hz - Loud",
                    start_time=05.4, file_start_time=05.4,
                    end_time=07.9, file_end_time=07.9),
        SegmentData(id=4, title="330Hz - Soft",
                    start_time=07.9, file_start_time=07.9,
                    end_time=10.4, file_end_time=10.4),
        SegmentData(id=5, title="440Hz - Loud",
                    start_time=10.4, file_start_time=10.4,
                    end_time=12.9, file_end_time=12.9),
        SegmentData(id=6, title="440Hz - Soft",
                    start_time=12.9, file_start_time=12.9,
                    end_time=16.399, file_end_time=16.399),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)


def test_slide_overshoot_forward(tmp_path, chaptered_audio_file_path):
    """Slide some chapters over the forward edge."""
    cmd = [str(chaptered_audio_file_path), "-d", "12.5"]
    expected = [
        SegmentData(id=0, title="110Hz - Loud",
                    start_time=00.0, file_start_time=00.0,
                    end_time=15.0, file_end_time=15.0),
        SegmentData(id=1, title="110Hz - Soft",
                    start_time=15.0, file_start_time=15.0,
                    end_time=17.5, file_end_time=17.5),
        SegmentData(id=2, title="220Hz - Loud",
                    start_time=17.5, file_start_time=17.5,
                    end_time=19.999, file_end_time=19.999),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)


def test_slide_overshoot_backwards(tmp_path, chaptered_audio_file_path):
    """Slide some chapters over the rear edge."""
    cmd = [str(chaptered_audio_file_path), "-d", "-12.5"]
    expected = [
        SegmentData(id=0, title="330Hz - Soft",
                    start_time=00.0, file_start_time=00.0,
                    end_time=02.5, file_end_time=02.5),
        SegmentData(id=1, title="440Hz - Loud",
                    start_time=02.5, file_start_time=02.5,
                    end_time=05.0, file_end_time=05.0),
        SegmentData(id=2, title="440Hz - Soft",
                    start_time=05.0, file_start_time=05.0,
                    end_time=19.999, file_end_time=19.999),
    ]
    _slide_and_check_segments(cmd, chaptered_audio_file_path, expected)



from unittest import mock

import pytest
import testhelpers

from m4b_util.helpers import ffprobe, SegmentData
from m4b_util.subcommands import labels


@pytest.fixture
def label_file_path(tmp_path):
    """A generic label file."""
    label_file = tmp_path / "labels.txt"
    with open(label_file, "w") as f:
        f.write("0.000000	0.000000	1 - 110Hz" "\n"
                "2.500000	2.500000	2 - 220Hz" "\n"
                "5.000000	5.000000	3 - 330Hz" "\n"
                "7.500000	7.500000	4 - 440Hz" "\n"
                "10.000000	10.000000	5 - 550Hz" "\n"
                "12.500000	12.500000	6 - 660Hz" "\n"
                "15.000000	15.000000	7 - 770Hz" "\n"
                "17.600000	19.999000	8 - 880Hz" "\n"
                )
    return label_file


def _run_labels_cmd(arg_list):
    """Patch the runtime arguments, then run the split command."""
    argv_patch = ["m4b-util", "labels"]
    argv_patch.extend(arg_list)

    with mock.patch("sys.argv", argv_patch):
        labels.run()


def test_segment_data_from_labels():
    """Convert labels to a list of segment data."""
    input_list = [
        "0.000000	0.000000	Start",
        "75.223314	75.223314	Segment 1",
        "186.845005	186.845005	Segment 2",
        "320.305723	320.305723	Segment 3",
        "408.632453	408.632453	Segment 4",
        "461.531429	461.531429	End"
    ]
    expected = [
        SegmentData(start_time=000.000, end_time=075.223, title="Start"),
        SegmentData(start_time=075.223, end_time=186.845, title="Segment 1"),
        SegmentData(start_time=186.845, end_time=320.306, title="Segment 2"),
        SegmentData(start_time=320.306, end_time=408.632, title="Segment 3"),
        SegmentData(start_time=408.632, end_time=461.531, title="Segment 4"),
        SegmentData(start_time=461.531, end_time=461.531, title="End"),
    ]
    actual = labels.segment_data_from_labels(input_list)
    assert actual == expected


def test_segment_data_from_labels_invalid_labels(capsys):
    """Convert labels to a list of segment data."""
    input_list = [
        "0.0000.00	0.000000	Start",
        "0.000000	0.000.000	StartAgain",
        "75.223314	75.223314	Segment 1",
        "186.845005	186.845005	Segment 2",
        "derp",
        "320.305723	320.305723	Segment 3",
        "408.632453	408.632453	",
        "461531429	461.531429	End"
    ]
    expected = [
        SegmentData(start_time=075.223, end_time=186.845, title="Segment 1"),
        SegmentData(start_time=186.845, end_time=320.306, title="Segment 2"),
        SegmentData(start_time=320.306, end_time=408.632, title="Segment 3"),
        SegmentData(start_time=408.632, end_time=461531429., title=""),
        SegmentData(start_time=461531429., end_time=461.531, title="End"),
    ]
    actual = labels.segment_data_from_labels(input_list)
    assert actual == expected

    output = capsys.readouterr()
    for expected in ["Could not parse label:", "Start", "StartAgain", "derp"]:
        assert expected in output.out


def test_labels_from_segment_data():
    """Create a list of labels from a list of segment data."""
    input_list = [
        SegmentData(start_time=000.000, end_time=075.223, title="Start"),
        SegmentData(start_time=075.223, end_time=186.845, title="Segment 1"),
        SegmentData(start_time=186.845, end_time=320.306, title="Segment 2"),
        SegmentData(start_time=320.306, end_time=408.632),
        SegmentData(start_time=408.632, end_time=461.531, title=""),
        SegmentData(start_time=461.531, end_time=461.531, title="End"),
    ]
    expected = [
        "0.0	0.0	Start",
        "75.223	75.223	Segment 1",
        "186.845	186.845	Segment 2",
        "320.306	320.306	None",
        "408.632	408.632	",
        "461.531	461.531	End"
    ]
    actual = labels.labels_from_segment_data(input_list)
    assert actual == expected


def test_labels_from_labels(tmp_path, label_file_path, variable_volume_segments_file_path):
    """Generate all possible outputs from the label command, using a label file as input."""
    meta_file_path = tmp_path / "ffmetadata"
    label_out_path = tmp_path / "labels.out.txt"
    _run_labels_cmd([
        "--from-label-file", str(label_file_path),
        "--to-metadata-file", str(meta_file_path),
        "--to-label-file", str(label_out_path),
        "--to-book", str(variable_volume_segments_file_path)
    ])

    # Check metadata file
    expected_metadata = (
        ";FFMETADATA1\n"
        "major_brand=M4A\n"
        "minor_version=512\n"
        "compatible_brands=M4A isomiso2\n"
        "title=None\n"
        "artist=None\n"
        "album=None\n"
        "date=None\n"
        "genre=Audiobook\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=0\n"           "END=2500\n"        "title=1 - 110Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=2500\n"        "END=5000\n"        "title=2 - 220Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=5000\n"        "END=7500\n"        "title=3 - 330Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=7500\n"        "END=10000\n"       "title=4 - 440Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=10000\n"       "END=12500\n"       "title=5 - 550Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=12500\n"       "END=15000\n"       "title=6 - 660Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=15000\n"       "END=17600\n"       "title=7 - 770Hz\n"
        "[CHAPTER]\n"        "TIMEBASE=1/1000\n"        "START=17600\n"       "END=19999\n"       "title=8 - 880Hz\n"
    )
    with open(meta_file_path) as f:
        metadata = f.read()
    assert metadata == expected_metadata

    # Check Label file
    expected_labeldata = (
        "0.0	0.0	1 - 110Hz" "\n"
        "2.5	2.5	2 - 220Hz" "\n"
        "5.0	5.0	3 - 330Hz" "\n"
        "7.5	7.5	4 - 440Hz" "\n"
        "10.0	10.0	5 - 550Hz" "\n"
        "12.5	12.5	6 - 660Hz" "\n"
        "15.0	15.0	7 - 770Hz" "\n"
        "17.6	17.6	8 - 880Hz" "\n"
    )
    with open(label_out_path) as f:
        labeldata = f.read()
    assert labeldata == expected_labeldata

    # Check audiobook output
    probe = ffprobe.run_probe(variable_volume_segments_file_path)
    assert probe and len(probe.chapters) == 8
    assert probe.chapters[0]["tags"]["title"] == "1 - 110Hz"
    assert probe.chapters[1]["tags"]["title"] == "2 - 220Hz"
    assert probe.chapters[2]["tags"]["title"] == "3 - 330Hz"
    assert probe.chapters[3]["tags"]["title"] == "4 - 440Hz"
    assert probe.chapters[4]["tags"]["title"] == "5 - 550Hz"
    assert probe.chapters[5]["tags"]["title"] == "6 - 660Hz"
    assert probe.chapters[6]["tags"]["title"] == "7 - 770Hz"
    assert probe.chapters[7]["tags"]["title"] == "8 - 880Hz"


def test_labels_from_book(tmp_path, chaptered_audio_file_path, expected_data, variable_volume_segments_file_path):
    """Generate all possible outputs from the label command, using an audiobook file as input."""
    meta_file_path = tmp_path / "ffmetadata"
    label_out_path = tmp_path / "labels.out.txt"
    _run_labels_cmd([
        "--from-book", str(chaptered_audio_file_path),
        "--to-metadata-file", str(meta_file_path),
        "--to-label-file", str(label_out_path),
        "--to-book", str(variable_volume_segments_file_path)
    ])

    # Check metadata file
    with open(meta_file_path) as f:
        metadata = f.read()
    assert metadata == expected_data["chaptered"]["metadata"]

    # Check Label file
    expected_labeldata = (
        "0.0	0.0	110Hz - Loud" "\n"
        "2.5	2.5	110Hz - Soft" "\n"
        "5.0	5.0	220Hz - Loud" "\n"
        "7.5	7.5	220Hz - Soft" "\n"
        "10.0	10.0	330Hz - Loud" "\n"
        "12.5	12.5	330Hz - Soft" "\n"
        "15.0	15.0	440Hz - Loud" "\n"
        "17.5	17.5	440Hz - Soft" "\n"
    )
    with open(label_out_path) as f:
        labeldata = f.read()
    assert labeldata == expected_labeldata

    # Check audiobook output
    probe = ffprobe.run_probe(variable_volume_segments_file_path)
    assert probe and len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "110Hz - Loud"
    assert probe.chapters[1]['tags']['title'] == "110Hz - Soft"
    assert probe.chapters[2]['tags']['title'] == "220Hz - Loud"
    assert probe.chapters[3]['tags']['title'] == "220Hz - Soft"
    assert probe.chapters[4]['tags']['title'] == "330Hz - Loud"
    assert probe.chapters[5]['tags']['title'] == "330Hz - Soft"
    assert probe.chapters[6]['tags']['title'] == "440Hz - Loud"
    assert probe.chapters[7]['tags']['title'] == "440Hz - Soft"


def test_labels_overshoot(label_file_path, abrupt_ending_file_path):
    """Trim labels that extend beyond file when writing to book. Truncate those that begin inside, but end outside."""""
    _run_labels_cmd([
        "--from-label-file", str(label_file_path),
        "--to-book", str(abrupt_ending_file_path)
    ])
    # Check audiobook output
    probe = ffprobe.run_probe(abrupt_ending_file_path)
    assert probe and len(probe.chapters) == 7
    assert probe.chapters[0]['tags']['title'] == "1 - 110Hz"
    assert probe.chapters[1]['tags']['title'] == "2 - 220Hz"
    assert probe.chapters[2]['tags']['title'] == "3 - 330Hz"
    assert probe.chapters[3]['tags']['title'] == "4 - 440Hz"
    assert probe.chapters[4]['tags']['title'] == "5 - 550Hz"
    assert probe.chapters[5]['tags']['title'] == "6 - 660Hz"
    assert probe.chapters[6]['tags']['title'] == "7 - 770Hz"
    assert probe.chapters[6]['end_time'] == "17.500000"


def test_labels_no_output(label_file_path, tmp_path):
    """Don't write any files if no output is specified."""
    _run_labels_cmd(["--from-label-file", str(label_file_path)])
    expected_files = [label_file_path.name]
    testhelpers.check_output_folder(tmp_path, expected_files, check_func=testhelpers.assert_file_path_is_file)


def test_from_metadata(tmp_path, chaptered_audio_file_path, expected_data):
    """Read and apply metadata from a file."""
    # Write the input metadata to a file
    metadata_file_path = tmp_path / "ffmetadata"
    metadata = r"""
        ;FFMETADATA1
        major_brand=M4A
        minor_version=512
        compatible_brands=M4A isomiso2
        title=Super Secret: The Title
        artist=A Great Author
        album=Super Great Album
        date=2011
        genre=Audiobook
        [CHAPTER]
        TIMEBASE=1/1000
        START=0
        END=15000
        title=Opening Credits
        [CHAPTER]
        TIMEBASE=1/1000
        START=330987
        title=Not Included
        """
    with open(metadata_file_path, "w") as f:
        f.write(metadata)

    # Generate the output files
    label_out_path = tmp_path / "labels.out.txt"
    _run_labels_cmd([
        "--from-metadata-file", str(metadata_file_path),
        "--to-label-file", str(label_out_path),
        "--to-book", str(chaptered_audio_file_path)
    ])

    # Check Label file
    expected_labeldata = (
        "0.0	0.0	Opening Credits" "\n"
    )
    with open(label_out_path) as f:
        labeldata = f.read()
    assert labeldata == expected_labeldata

    # Check audiobook output
    probe = ffprobe.run_probe(chaptered_audio_file_path)
    assert probe and len(probe.chapters) == 1
    assert probe.chapters[0]['tags']['title'] == "Opening Credits"


def test_bad_metadata(tmp_path, capsys):
    """Handle bad metadata input."""
    metadata_file_path = tmp_path / "ffmetadata"
    metadata = r"""
        """
    with open(metadata_file_path, "w") as f:
        f.write(metadata)

    _run_labels_cmd(["--from-metadata-file", str(metadata_file_path)])

    output = capsys.readouterr()
    assert "Parsing metadata failed" in output.out



# Standard Library
from dataclasses import dataclass, field
from pathlib import Path
import shutil
from tempfile import mkdtemp

# Third Party
from lark import Lark
from natsort import natsorted
from rich import print
from rich.status import Status

# Local
from . import cover_utils, ffprobe, ffprogress
from .ffmetadata import FFMETADATA_GRAMMAR, FFMetadataTransformer
from .finders import find_chapters
from .parallel_ffmpeg import ParallelFFmpeg
from .segment_data import SegmentData


@dataclass
class Audiobook:
    """Store all the info pertaining to an audiobook."""
    author: str = None
    # Set duration to a big number, (hopefully larger than our actual duration)
    # just in case we try to write chapters without knowing the final duration.
    chapters: list = field(default_factory=lambda: [])
    cover: str = None
    date: str = None
    output_name: str = None
    title: str = None
    keep_temp_files: bool = False
    _tmp_dir: Path = field(init=False, repr=False, default=None)

    @property
    def _tmp_path(self):
        """Only create our temporary directory if we need it."""
        if not self._tmp_dir:
            self._tmp_dir = Path(mkdtemp(suffix="m4b"))
        if not self._tmp_dir.exists():
            self._tmp_dir.mkdir()
        return self._tmp_dir

    @property
    def suggested_file_name(self):
        """Calculate output name based on what we currently know."""
        # Use a default is self.output_name is not set.
        name = self.output_name or f"{self.author} - {self.title}.m4b"

        # Make sure it ends with .m4b
        if not name.endswith(".m4b"):
            name = f"{name}.m4b"
        return name

    @property
    def metadata(self):
        """Generate the metadata text when requested, since it depends on the current state of the object."""
        # Start by writing the global header
        metadata = (";FFMETADATA1\n"
                    "major_brand=M4A\n"
                    "minor_version=512\n"
                    "compatible_brands=M4A isomiso2\n"
                    f"title={self.title}\n"
                    f"artist={self.author}\n"
                    f"album={self.title}\n"
                    f"date={self.date}\n"
                    "genre=Audiobook\n"
                    )
        # Then add chapters to metadata
        for chapter in self.chapters:
            title = chapter.title
            # Convert times to miliseconds
            start = int(chapter.start_time * 1000)
            end = int(chapter.end_time * 1000)
            metadata += ("[CHAPTER]\n"
                         "TIMEBASE=1/1000\n"
                         f"START={start}\n"
                         f"END={end}\n"
                         f"title={title}\n"
                         )
        return metadata

    @metadata.setter
    def metadata(self, value):
        """Set metadata from a string."""
        parser = Lark(FFMETADATA_GRAMMAR, start='metadata', parser='lalr', transformer=FFMetadataTransformer())
        metadata = parser.parse(value)
        self.title = metadata['global'].get('title')
        self.author = metadata['global'].get('artist')
        self.date = metadata['global'].get('date')
        for chapter in metadata['chapters']:
            # If we don't have a start and end time, we can't use this chapter, so skip it.
            if chapter.get("START") and chapter.get("END"):
                # TIMEBASE is a fraction represented as a string. In order to use it, we will need to convert it to
                # a numerical value.
                timebase = chapter.get("TIMEBASE", "1/1000")  # Default to 1ms
                num, den = timebase.split('/')
                timebase = int(num) / int(den)

                # Put it all together!
                self.chapters.append(SegmentData(
                    title=chapter.get('title'),
                    start_time=int(chapter.get('START')) * timebase,
                    end_time=int(chapter.get('END')) * timebase
                ))

    @staticmethod
    def scan_dir(input_dir):
        """Scan a directory for files."""
        return natsorted(Path(input_dir).glob("*"))

    def add_chapters_from_chaptered_file(self, input_path):
        """Add chapters from a single file."""
        # Read the new chapters, shifting the times forward to match any segments we already have.
        time_shift = 0.0
        id_shift = 0
        if self.chapters:
            time_shift = self.chapters[-1].end_time
            id_shift = self.chapters[-1].id + 1
        new_chapters = find_chapters(input_path)
        for chapter in new_chapters:
            chapter.start_time = chapter.start_time + time_shift
            chapter.end_time = chapter.end_time + time_shift
            chapter.id = chapter.id + id_shift

        self.chapters.extend(new_chapters)
        # Run ffprobe and parse the output.
        probe = ffprobe.run_probe(input_path)
        if not probe or probe.audio is None:
            print(f"[bold yellow]Warning:[/] Unable to parse '[bold white]{input_path}[/]'. Skipping.")
            return

        # Attempt to fill any missing metadata
        if probe.tags:  # pragma: no branch - None of our test files hit this, but some other audio formats might.
            self.title = self.title or probe.tags.get('title')
            self.author = self.author or probe.tags.get('artist')
            self.date = self.date or probe.tags.get('date')

    def add_chapters_from_directory(self, input_dir, use_filenames=False, decode_durations=False):
        """Read files from a directory that represents an audiobook.

        :param input_dir: Directory to scan.
        :param use_filenames: Use filenames as chapter titles, instead of embedded metadata.
        :param decode_durations: Find duration by fully decoding file, instead of relying on metadata.
        """
        return self.add_chapters_from_filelist(self.scan_dir(input_dir), use_filenames, decode_durations)

    def add_chapters_from_filelist(self, input_files, use_filenames=False, decode_durations=False):
        """Read files from a list of files that represent an audiobook.

        :param input_files: Directory to scan.
        :param use_filenames: Use filenames as chapter titles, instead of embedded metadata.
        :param decode_durations: Find duration by fully decoding file, instead of relying on metadata.
        """
        # Set up variables
        time_counter = 0.0
        if self.chapters:  # If we already have chapters, add these on to the end.
            time_counter = self.chapters[-1].end_time
        segment_counter = len(self.chapters)

        # Start our status tracker
        print("[cyan]Collecting file data...[/]")
        file_scan_status = Status("Starting")
        file_scan_status.start()

        # Scan all the files
        for file in input_files:
            file_scan_status.update(f"Scanning {file.name}")

            # Check for cover file
            if file.stem == "cover" and file.suffix.lower() in [".png", ".jpg", ".jpeg"]:
                self.cover = file
                continue

            # Run ffprobe and parse the output.
            probe = ffprobe.run_probe(file)
            if not probe or probe.audio is None:
                print(f"[bold yellow]Warning:[/] Unable to parse '[bold white]{file}[/]'. Skipping.")
                continue

            # Count which segment we are on. Used in case the chapter has no title.
            segment_counter += 1

            # Attempt to fill any missing metadata
            chapter_title = None
            if probe.tags:  # pragma: no branch - None of our test files hit this, but some other audio formats might.
                self.title = self.title or probe.tags.get('album')
                self.author = self.author or probe.tags.get('artist')
                self.date = self.date or probe.tags.get('date')
                chapter_title = probe.tags.get('title')

            if use_filenames:
                chapter_title = file.stem
            if not chapter_title:
                chapter_title = str(segment_counter)

            # Get the duration, and calculate start/end times.
            start_time = time_counter
            try:
                duration = ffprobe.get_file_duration(file, decode_durations)
            except RuntimeError:
                print(f"[yellow]Warning:[/] Failed to determine duration of '[bold white]{file}[/]'. Ignoring.")
                continue
            time_counter += duration  # Update for next iteration
            end_time = time_counter

            # Record the chapter
            self.chapters.append(SegmentData(
                id=segment_counter,
                title=chapter_title,
                start_time=start_time,
                end_time=end_time,
                backing_file=file
            ))

        # Shut off the status tracker and alert the user
        file_scan_status.stop()
        print("[green]File scan complete.")

    def bind(self, output_path):
        """Bind together an audiobook, based on what we know."""
        if self.keep_temp_files:
            print(f"[yellow]Info:[/] Temp files can be found at [yellow]{self._tmp_path}[/]")

        # Make sure we have at least one chapter.
        if len(self.chapters) == 0:
            print("[bold red]Error:[/] Nothing to bind.")
            return False

        # Determine if we need to bind multiple files, or just rewrite the metadata of a single file.
        first_file = None
        unaccounted_duration = 999  # Will be overwritten, but set to a positive number just in case.
        for segment in self.chapters:
            if segment.backing_file is None:
                s_id = segment.id or ""
                print(f"[yellow]Info:[/] Segment {s_id}: '{segment.title}' does not point to a file.")
                print("[bold red]Error:[/] Cannot bind a non-backed segment.")
                return False

            # Keep info on the first file we see
            if not first_file:
                first_file = segment.backing_file
                unaccounted_duration = ffprobe.get_file_duration(first_file)

            # If we see a file different from the first one, switch to multi-segment bind mode
            if segment.backing_file != first_file:
                return self._bind_multiple_segments(output_path)

            # Subtract this file's duration from our unaccounted_duration metric
            unaccounted_duration = unaccounted_duration - (segment.end_time - segment.start_time)

        # Check to see if we accounted for the total duration of the file. If not, we must be using only part of the
        # original file, and will need to switch to multi-segment bind.
        if unaccounted_duration > 0.1:
            return self._bind_multiple_segments(output_path)

        # If our segments cover the entirety of a single file, we can assume we just need to rewrite the metadata.
        out_file = self._add_chapter_info(first_file)
        return self._finish_bind(out_file, output_path)

    def _bind_multiple_segments(self, output_path):
        # First, convert all audios to m4a, in parallel.
        p = ParallelFFmpeg("Converting files to m4a")
        tasks = list()
        temp_files = list()
        finished_file = self._tmp_path / "finished.m4b"
        for i, segment in enumerate(self.chapters):
            file = segment.backing_file
            # Determine what to call the output file
            out_m4a = self._tmp_path / f"{i}_{file.stem}.m4a"

            # Record the file for concatenation later
            temp_files.append(out_m4a)

            # Write the FFMPEG command
            cmd = ["ffmpeg"]
            if segment.file_start_time:
                cmd.extend(["-ss", str(segment.file_start_time)])
            cmd.extend(["-i", file])
            if segment.file_end_time:
                start_time = segment.file_start_time or 0.0
                end_time_dur = segment.file_end_time - start_time
                cmd.extend(["-t", str(end_time_dur)])
            if segment.title:
                cmd.extend(["-metadata", f"title={segment.title}"])
            cmd.extend(["-filter_complex", "[0:a]asetpts=N/SR/TB[s0]", "-map", "[s0]",
                        "-c:a", "aac", out_m4a, "-y"])
            tasks.append({
                "name": file.stem,
                "command": cmd
            })
        p.process(tasks)

        # Save our original chapter info, so we can restore it later.
        old_chapters = self.chapters
        # Scan the files we just converted, to make sure the metadata matches
        self.chapters = list()
        self.add_chapters_from_filelist(temp_files)

        # Next, concatenate those m4a files into an m4b
        long_file = self._concatenate_files(temp_files)

        # Now, add chapter info
        coverless_file = self._add_chapter_info(long_file)

        # Restore the old chapter list, as we are done writing metadata
        self.chapters = old_chapters

        # Add cover artwork if specified, otherwise just copy the previous output
        if self.cover:
            covered_file = self._tmp_path / "covered.m4b"
            try:
                cover_utils.add_cover(coverless_file, self.cover, covered_file)
            except RuntimeError as e:
                print(str(e))
                exit(1)
            shutil.move(covered_file, finished_file)
        else:
            shutil.move(coverless_file, finished_file)

        return self._finish_bind(finished_file, output_path)

    def _finish_bind(self, input_path, output_path):
        """Copy the output and cleanup."""
        # Copy the output into place
        print("[cyan]Copying output.")
        shutil.copy(input_path, output_path)

        # Clean up
        if not self.keep_temp_files:
            shutil.rmtree(self._tmp_path)

        return True

    def _concatenate_files(self, input_list):
        """Bring all the input files, now m4a, and put them into an m4b."""
        # Start by writing the filelist to a temp file.
        filelist_file = self._tmp_path / "filelist"
        with open(filelist_file, 'w') as f:
            for file in input_list:
                # Single quotes in filenames need to be replaced with `'\''` to work with ffmpeg
                # https://superuser.com/questions/787064/filename-quoting-in-ffmpeg-concat
                safer_filename = str(file).replace("'", "'\\''")
                f.write(f"file '{safer_filename}'\n")

        # Then use that file to concatenate everything.
        output_file = self._tmp_path / "long.m4a"
        cmd = ["ffmpeg", "-f", "concat", "-safe", "0", "-i", filelist_file, "-c", "copy", output_file]
        self._run_ffmpeg(cmd, "Combining Audio Files.")

        return output_file

    def _add_chapter_info(self, input_file):
        """Add chapter info to an m4b file."""
        # Write the metadata to a temp file.
        metadata_file = self._tmp_path / "ffmetadata"
        with open(metadata_file, 'w') as f:
            print(self.metadata, file=f)

        output_file = self._tmp_path / "coverless.m4b"
        cmd = ["ffmpeg", "-y", "-i", input_file, "-i", metadata_file, "-map_metadata", "1", "-map_chapters", "1",
               "-c", "copy", output_file]
        self._run_ffmpeg(cmd, "Converting to .m4b")

        return output_file

    def _run_ffmpeg(self, cmd, message):
        if not ffprogress.run(cmd, message):
            print("[bold red]Error:[/] ffmpeg failed.")
            if not self.keep_temp_files:
                shutil.rmtree(self._tmp_path)
            exit(1)



# Standard Library
import argparse
from pathlib import Path
import re
import sys

# Third Party
from lark import UnexpectedInput
from rich import print

# Local
from ..helpers import Audiobook, ffprobe, SegmentData
from ..helpers.ffmetadata import FFMETADATA_TERMINATOR_FRIENDLY_NAMES


def _parse_args():
    parser = argparse.ArgumentParser(
        prog="m4b-util load-labels",
        description="Convert between audacity labels and ffmpeg chapter metadata."
    )
    # Inputs
    input_options = parser.add_mutually_exclusive_group(required=True)
    input_options.add_argument("--from-book", help="Read chapters from file.")
    input_options.add_argument("--from-label-file", help="Read audacity labels from text file.")
    input_options.add_argument("--from-metadata-file", help="Read ffmpeg metadata from file.")

    # Outputs
    output_options = parser.add_argument_group("output options")
    output_options.add_argument("--to-metadata-file", type=str, help="Output ffmpeg metadata to file.")
    output_options.add_argument("--to-label-file", type=str, help="Output labels to file.")
    output_options.add_argument("--to-book", type=str, help="Apply labels as chapters to existing book file.")

    args = parser.parse_args(sys.argv[2:])
    return args


def segment_data_from_labels(labels):
    """Read labels from a file and fill a segment data list.

    :param labels: A list of labels lines, preferably from running readlines() on a label file.
    :returns: A list of SegmentData
    """
    segments = list()
    label_regex = re.compile(r"^(?P<start_time>\d+\.?\d*)\s+(?P<end_time>\d+\.?\d*)\s+(?P<title>.*)\s*$")
    previous = None
    for label in labels:
        match = label_regex.search(label)
        if match:
            if previous:
                segments.append(
                    SegmentData(
                        start_time=previous['start_time'],
                        end_time=float(match['start_time']),
                        title=previous['title'])
                )
            previous = {
                "start_time": float(match['start_time']),
                "end_time": float(match['end_time']),
                "title": match['title']
            }
        else:
            print(f"[orange]Warning:[/] Could not parse label: '{label}'")

    # Make sure to hande the final label
    segments.append(SegmentData(
        start_time=previous['start_time'],
        end_time=previous['end_time'],
        title=previous['title']
    ))

    return segments


def labels_from_segment_data(segments):
    """Generate labels from a segment data list."""
    labels = list()
    for segment in segments:
        labels.append(f"{segment.start_time}\t{segment.start_time}\t{segment.title}")
    return labels


def _handle_input(args, book):
    # Handle input
    if args.from_label_file:
        with open(args.from_label_file) as f:
            labels = f.readlines()
        book.chapters = segment_data_from_labels(labels)
    elif args.from_book:
        book.add_chapters_from_chaptered_file(args.from_book)
    else:  # args.from_metadata_file:
        with open(args.from_metadata_file) as f:
            metadata = f.read()
            try:
                book.metadata = metadata
            except UnexpectedInput as e:
                print(f"[red]Error:[/] Parsing metadata failed on line {e.line}, column {e.column}:")
                print(f"{e.get_context(metadata)}")
                print("Expected one of the following token types:")
                for terminator in e.accepts:
                    print(f" - {FFMETADATA_TERMINATOR_FRIENDLY_NAMES.get(terminator, terminator)}")
                return False
    return True


def _handle_output(args, book):
    if args.to_label_file:
        with open(args.to_label_file, 'w') as f:
            for label in labels_from_segment_data(book.chapters):
                f.write(f"{label}\n")
    if args.to_metadata_file:
        with open(args.to_metadata_file, 'w') as f:
            f.write(book.metadata)
    if args.to_book:
        new_book = Audiobook()
        new_book.add_chapters_from_chaptered_file(args.to_book)
        new_book_dur = ffprobe.get_file_duration(args.to_book)
        new_chapters = list()
        for chapter in book.chapters:
            # Check to see if a chapter is straddling the cutoff. If so, set the chapter's end equal to the cutoff.
            if chapter.end_time > new_book_dur > chapter.start_time:
                chapter.end_time = new_book_dur

            # Only add chapters that are within the new books bounds
            if chapter.end_time <= new_book_dur:
                new_chapters.append(
                    SegmentData(
                        start_time=chapter.start_time,
                        end_time=chapter.end_time,
                        title=chapter.title,
                        backing_file=Path(args.to_book),
                        file_start_time=chapter.start_time,
                        file_end_time=chapter.end_time
                    )
                )
        new_book.chapters = new_chapters
        new_book.bind(args.to_book)


def run():
    """Run the subcommand."""
    # Set up variables
    args = _parse_args()
    book = Audiobook()

    # Handle input
    if not _handle_input(args, book):
        return 1

    # Handle output
    _handle_output(args, book)



"""Helper functions for running ffprobe."""
import json
import re
from subprocess import run


class Probe:
    """Holds the results from an ffprobe run."""
    def __init__(self, ffprobe_output):
        """Set up attributes."""
        self.data = json.loads(ffprobe_output)
        self.streams = self.data.get('streams', None)
        self.audio = self.first_stream_of(codec_type="audio")
        self.format = self.data.get('format')
        self.tags = self.data.get('format', dict()).get('tags', None)
        self.chapters = self.data.get('chapters')

    def first_stream_of(self, codec_type):
        """Get the first stream matching the requested arguments.

        For now, "type" is the only thing we care about, but maybe in the future we will want to filter on other fields?

        :param codec_type: Stream type. Usually "audio" or "video".
        :return: List representing the first matching stream
        """
        for stream in self.data.get('streams'):
            if stream.get("codec_type") == codec_type:
                return stream

        # No stream matches
        return None


def run_probe(file):
    """Run ffprobe on the specified file and return a JSON representation of the output."""
    cmd = ["ffprobe", "-show_format", "-show_streams", "-show_chapters", "-of", "json", "-i", file]

    p = run(cmd, capture_output=True)
    if p.returncode != 0:
        return None
    return Probe(p.stdout.decode('utf-8'))


def get_file_duration(file, decode_duration=False):
    """Determine the duration of a file.

    If self.decode_durations is set, run the file through ffmpeg to get the time.
    Otherwise, get it from the metadata.

    :param file: The audio file to run through ffmpeg
    :param decode_duration: Actually decode the file to find duration, instead of relying on metadata.

    :return The duration of the file, in seconds
    """
    duration = None
    if decode_duration:
        # Setup variables
        time_regex = re.compile(r"time=(?P<hour>\d{2}):(?P<min>\d{2}):(?P<sec>\d{2})\.(?P<ms>\d{2})")
        time_match = None

        # Run the decoder
        cmd = ["ffmpeg", "-i", file, "-loglevel", "info", "-nostats", "-f", "null", "-"]
        p = run(cmd, capture_output=True)

        # Search for time messages
        for line in p.stderr.decode('utf-8', errors="replace").splitlines():
            match = time_regex.search(line)
            if match:
                time_match = match  # Only keep the most recent

        # Set the duration if we found any
        if time_match:
            match = time_match.groupdict()
            duration = ((int(match["hour"]) * 60 * 60)
                        + (int(match["min"]) * 60)
                        + int(match["sec"])
                        + (int(match["ms"]) / 100))
        else:
            print(f"[bold yellow]Warning:[/] Unable to find duration of '[bold white]{file.name}[/]' "
                  "during decoding. Falling back to metadata.")

    if not duration:  # Use the metadata as given
        probe = run_probe(file)
        if not probe or probe.audio is None:
            raise RuntimeError("Could not get audio stream.")
        try:
            seconds = float(probe.audio.get('duration'))
        except TypeError:
            raise RuntimeError("Cannot parse duration listed in file.")
        duration = seconds

    return duration



import argparse
from pathlib import Path
import sys

from rich import print

from ..helpers import Audiobook


def _parse_args():
    parser = argparse.ArgumentParser(
        prog="m4b-util slide",
        description='Slide chapter segments up or down.'
    )
    parser.add_argument('input_file', help='Input filename')
    parser.add_argument('-d', "--duration", type=float, help='Duration to shift (seconds). Negative values allowed.')
    parser.add_argument("--trim-start", type=float, help='Duration to trim from beginning (seconds).')

    return parser.parse_args(sys.argv[2:])


def _slide(segments, duration):
    """Add duration to all relevant fields each segment."""
    for segment in segments:
        segment.start_time = segment.start_time + duration
        segment.end_time = segment.end_time + duration
        if segment.file_start_time is not None:
            segment.file_start_time = segment.file_start_time + duration
        if segment.file_end_time is not None:
            segment.file_end_time = segment.file_end_time + duration


def _check_overshoot(segments, start_time, end_time, file_start_time, file_end_time):
    """Remove any segments that exist outside the start and end time."""
    for segment in segments.copy():
        if segment.start_time > end_time \
                or segment.end_time <= 0 \
                or segment.end_time < start_time:
            segments.remove(segment)
            continue
        # Don't do any operations on backing_file info if one of the needed vars is None.
        if None not in (segment.file_start_time, segment.file_end_time, file_start_time, file_end_time):
            if segment.file_start_time > file_end_time \
                    or segment.file_end_time <= 0 \
                    or segment.file_end_time < file_start_time:
                segments.remove(segment)


def _slide_segments(segments, duration):
    """Slide all segments in a list in the indicated direction. Will not overshoot final or initial values.

    :param segments: List of segments to slide.
    :param duration: How far to slide the segments.
    :return: List of segments, slid by duration.
    """
    # Check preconditions
    if len(segments) == 0:
        return segments  # No values, so no slide
    if duration == 0:
        return segments  # No slide duration, no slide needed

    # Set initial state
    start_time = segments[0].start_time
    end_time = segments[-1].end_time
    duration = float(duration)
    file_start_time = segments[0].file_start_time
    file_end_time = segments[-1].file_end_time

    # Slide all segments
    _slide(segments, duration)

    # Check for overshoot
    original_length = len(segments)
    _check_overshoot(segments, start_time, end_time, file_start_time, file_end_time)

    # Renumber segments, if applicable
    if len(segments) != original_length:
        for i, segment in enumerate(segments):
            segment.id = i

    # Restore original start and end times, if we have any segments left.
    if segments:
        segments[-1].end_time = end_time
        segments[0].start_time = start_time
        # These will be None if unset
        segments[-1].file_end_time = file_end_time
        segments[0].file_start_time = file_start_time

    return segments


def run():
    """Split an audio file into pieces, based on silence."""
    args = _parse_args()
    input_path = Path(args.input_file)
    slide_dur = 0
    if args.duration:
        slide_dur = args.duration
    if args.trim_start:
        slide_dur = slide_dur - args.trim_start

    book = Audiobook()
    book.add_chapters_from_chaptered_file(input_path)
    if book.chapters:
        # Find initial end times, in case we need to modify it.
        end_time = book.chapters[-1].end_time
        file_end_time = book.chapters[-1].file_end_time

        if args.trim_start:
            # Remove chapters smaller than our trim value
            # Use .copy so that we can delete from the original without screwing up the iterator
            for chapter in book.chapters.copy():
                if chapter.end_time < args.trim_start:
                    book.chapters.remove(chapter)
                else:
                    break
            if not book.chapters:
                print("[bold red]Error:[/] No chapters were left after trim.")
                exit(1)
            # Set expected start and end times, since the slide function won't touch those
            book.chapters[0].start_time = 0.0
            book.chapters[0].file_start_time = 0.0
            book.chapters[-1].end_time = end_time - args.trim_start
            book.chapters[-1].file_end_time = file_end_time - args.trim_start

        # Slide any remaining chapters
        book.chapters = _slide_segments(book.chapters, slide_dur)
        if not book.chapters:
            print("[bold red]Error:[/] No chapters were left after slide.")
            exit(1)

        # Write our new file
        if book.bind(input_path) is False:  # pragma: nocover - failures here aren't likely the fault of this module
            exit(1)
        print(f"Chapters slid by {args.duration}")
        return 0

    else:
        print("[red]Error:[/] No chapters found!")
        exit(1)



"""Test the Bind subcommand."""
from unittest.mock import patch

import m4b_util


def _run_bind_cmd(arg_list):
    argv_patch = ["m4b-util", "bind"]
    argv_patch.extend(arg_list)

    with patch("sys.argv", argv_patch):
        m4b_util.subcommands.bind.run()


def test_show_order(mp3_path, capsys):
    """Show the order the files would be bound in, if asked."""
    _run_bind_cmd([str(mp3_path), "--show-order"])

    output = capsys.readouterr()
    expected_output = ('  0:    1 - 110Hz.mp3\n'
                       '  1:    2 - 220Hz.mp3\n'
                       '  2:    3 - 330Hz.mp3\n'
                       '  3:    4 - 440Hz.mp3\n'
                       '  4:    5 - 550Hz.mp3\n'
                       '  5:    6 - 660Hz.mp3\n'
                       '  6:    7 - 770Hz.mp3\n'
                       '  7:    8 - 880Hz.mp3\n'
                       )
    assert (output.out == expected_output)


def test_show_order_for_specific_files(mp3_path, capsys):
    """Show the order the files would be bound in, even when manually specified."""
    _run_bind_cmd(["-f" f"{mp3_path}/5 - 550Hz.mp3", "-f" f"{mp3_path}/1 - 110Hz.mp3", "--show-order"])

    output = capsys.readouterr()
    expected_output = ('  0:    5 - 550Hz.mp3\n'
                       '  1:    1 - 110Hz.mp3\n'
                       )
    assert (output.out == expected_output)


def test_nonexistent_output_dir(capsys):
    """Alert the user if the output dir doesn't exist."""
    _run_bind_cmd(["fake/input/dir", "--output-dir", "/some/nonexistent/directory/"])

    output = capsys.readouterr()
    assert ("is not a directory" in output.out)


def test_bind_wav_defaults(wav_path, tmp_path, capsys):
    """Bind a folder of wav files into and audiobook."""
    _run_bind_cmd([str(wav_path), "-o", str(tmp_path)])

    # Check for output
    output_path = tmp_path / "None - None.m4b"
    assert output_path.is_file()

    probe = m4b_util.helpers.ffprobe.run_probe(output_path)
    assert probe
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"

    # Make sure we tell the user where the output file is. Since rich auto-wraps the output unpredictably, we will
    # just check to make sure the start of the message, and the book file name are present.
    output = capsys.readouterr()
    assert ("Writing " in output.out)
    assert ("None - None.m4b" in output.out)


def test_bind_fail(tmp_path, capsys):
    """Avoid telling the user we output a file when we didn't."""
    fake_folder = tmp_path / "fake"
    fake_folder.mkdir()
    _run_bind_cmd([str(fake_folder)])
    output = capsys.readouterr()
    assert "Writing" not in output.out


def test_bind_use_filename(wav_path):
    """Use the filenames as the chapter titles."""
    with patch("m4b_util.subcommands.bind.Audiobook") as mock_book:
        mock_book = mock_book.return_value  # Required since we want to mock a specific instance, not the class.
        _run_bind_cmd([str(wav_path), "--use-filename"])
        # Since assert_called_with would require us to specify all arguments, we check the call args for the one we care
        # about manually.
        assert mock_book.add_chapters_from_directory.call_args.kwargs['use_filenames'] is True


def test_bind_specific_files(wav_path, tmp_path, capsys):
    """Bind a folder of wav files into and audiobook."""
    _run_bind_cmd(["-o", str(tmp_path), "-f", f"{wav_path}/1 - 110Hz.wav", "-f", f"{wav_path}/3 - 330Hz.wav"])

    # Check for output
    output_path = tmp_path / "None - None.m4b"
    assert output_path.is_file()

    probe = m4b_util.helpers.ffprobe.run_probe(output_path)
    assert probe
    assert probe.format['duration'] == "10.021333"
    assert probe.audio['duration'] == "10.021333"
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"


def test_bind_files_specified_with_input_dir_too(mp3_path, tmp_path, capsys):
    """Bind specific files, even if an input directory is specified."""
    _run_bind_cmd(
        [str(mp3_path), "-o", str(tmp_path), "-f", f"{mp3_path}/1 - 110Hz.mp3", "-f", f"{mp3_path}/3 - 330Hz.mp3"]
    )

    # Check that we warn the user.
    output = capsys.readouterr()
    assert ("Warning: Both an input folder and specific files were specified." in output.out)

    # Check for output
    output_path = tmp_path / "None - None.m4b"
    assert output_path.is_file()
    probe = m4b_util.helpers.ffprobe.run_probe(output_path)
    assert probe.format['duration'] == "10.021333"


def test_bind_no_files(capsys):
    """Alert the user if no files were specified."""
    _run_bind_cmd([])

    output = capsys.readouterr()
    assert ("You must provide either an input folder or a list of files" in output.out)



"""Test the Audiobook class."""
from pathlib import Path
import shutil
from subprocess import run
from tempfile import TemporaryDirectory

from natsort import natsorted
import testhelpers

from m4b_util.helpers import Audiobook, ffprobe, SegmentData


def _do_dir_scan(book, in_dir, use_filenames=False, decode_durations=False, **kwargs):
    """Read book from directory, then test the book's attributes based on kwargs."""
    book.add_chapters_from_directory(in_dir, use_filenames, decode_durations)
    _check_attrs(book, **kwargs)


def _do_file_scan(book, file_list, use_filenames=False, decode_durations=False, **kwargs):
    """Read book from a list of files, then test the book's attributes based on kwargs."""
    book.add_chapters_from_filelist(file_list, use_filenames, decode_durations)
    _check_attrs(book, **kwargs)


def _check_attrs(book, **kwargs):
    """Check book's attrs against the provided kwargs."""
    testhelpers.soften_backing_file(book.chapters)
    for attr_name, attr_value in kwargs.items():
        assert (getattr(book, attr_name) == attr_value)


def _add_metadata(file_path, **kwargs):
    """Add metadata specified by kwargs to file_path."""
    with TemporaryDirectory() as tmp_dir:
        tmp_path = Path(tmp_dir) / f"temp{file_path.suffix}"
        cmd = ["ffmpeg", "-i", file_path]

        # Add in metadata from kwargs
        for key, value in kwargs.items():
            cmd.append("-metadata")
            cmd.append(f"{key}={value}")

        # Add output path
        cmd.append(tmp_path)

        run(cmd, capture_output=True, check=True)

        shutil.move(tmp_path, file_path)


def test_suggested_file_name():
    """Set the output path in various ways."""
    b = Audiobook()
    assert (str(b.suggested_file_name) == "None - None.m4b")  # Default
    # Author
    b.author = "Write-y McAuthor"
    assert (str(b.suggested_file_name) == "Write-y McAuthor - None.m4b")
    # Author + Title
    b.title = "Test Book: An exercise of m4b-util"
    assert (str(b.suggested_file_name) == "Write-y McAuthor - Test Book: An exercise of m4b-util.m4b")
    # Name (Overwrites Author & Title), Adds m4b extension
    b.output_name = "Author McWrite-y"
    assert (str(b.suggested_file_name) == "Author McWrite-y.m4b")
    # Name (Overwrites Author & Title), does not double up m4b extension
    b.output_name = "Author McWrite-y.m4b"
    assert (str(b.suggested_file_name) == "Author McWrite-y.m4b")


def test_file_scanner_mp3(mp3_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    b = Audiobook()
    _do_dir_scan(
        b,
        mp3_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['mp3']['chapters'],
    )
    # Verify that scanning twice doubles the chapters
    _do_dir_scan(
        b,
        mp3_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['mp3']['chapters_doubled'],
    )


def test_file_scanner_wav(wav_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    b = Audiobook()
    _do_dir_scan(
        b,
        wav_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters'],
        metadata=expected_data['default']['metadata']
    )
    # Verify that scanning twice doubles the chapters
    _do_dir_scan(
        b,
        wav_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters_doubled'],
    )


def test_file_scanner_m4a(m4a_path, expected_data):
    """Get the correct data from a set of mp3 files."""
    b = Audiobook()
    _do_dir_scan(
        b,
        m4a_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters'],
        metadata=expected_data['default']['metadata']
    )
    # Verify that scanning twice doubles the chapters
    _do_dir_scan(
        b,
        m4a_path,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['default']['chapters_doubled'],
    )


def test_single_file_scanner(chaptered_audio_file_path, expected_data):
    """Get the correct data from a single, chaptered file."""
    b = Audiobook()
    b.add_chapters_from_chaptered_file(chaptered_audio_file_path)
    _check_attrs(
        b,
        title="Chaptered Audio",
        author="m4b-util",
        date="2022",
        chapters=expected_data['chaptered']['chapters'],
        metadata=expected_data['chaptered']['metadata']
    )
    b.add_chapters_from_chaptered_file(chaptered_audio_file_path)
    _check_attrs(b, chapters=expected_data['chaptered']['chapters_doubled'])


def test_file_scanner_use_filenames(mp3_path, expected_data):
    """Use filenames instead of metadata when scanning files."""
    b = Audiobook()
    _do_dir_scan(
        b,
        mp3_path,
        use_filenames=True,
        title=None,
        author=None,
        date=None,
        chapters=expected_data['use_filename']['chapters'],
    )


def test_metadata_from_input(mp3_file_path):
    """Pick up metadata from our input files."""
    _add_metadata(mp3_file_path, album="Book Title", artist="Author McWrite-y", date="2022")
    _do_file_scan(
        Audiobook(),
        [mp3_file_path],
        title="Book Title",
        author="Author McWrite-y",
        date="2022"
    )


def test_grab_first_metadata_only(mp3_path):
    """Only fill metadata items the first time we see them."""
    file = mp3_path / "2 - 220Hz.mp3"
    _add_metadata(file, album="Book Title", artist="Author McWrite-y")
    second_file = mp3_path / "3 - 330Hz.mp3"
    _add_metadata(second_file, album="Fake Book Title", artist="Not a Real Author")
    _do_dir_scan(
        Audiobook(),
        mp3_path,
        title="Book Title",
        author="Author McWrite-y"
    )


def test_prefer_user_supplied_metadata(mp3_file_path):
    """Don't fill metadata from files if we already have it."""
    _add_metadata(mp3_file_path, album="Fake Book Title", artist="Not Author McWrite-y", date="1011")
    b = Audiobook()
    b.title = "Book Title"
    b.author = "Author McWrite-y"
    b.date = "2022"
    _do_file_scan(
        b,
        [mp3_file_path],
        title="Book Title",
        author="Author McWrite-y",
        date="2022"
    )


def test_scanning_nonaudio_file(mp3_path, capsys):
    """Print a warning if we can't parse a file."""
    # Create a blank mp3 file
    open(mp3_path / "not-really-audio.mp3", 'a').close()
    _do_dir_scan(Audiobook(), mp3_path)
    output = capsys.readouterr()
    assert ("Warning: Unable to parse" in output.out)


def test_status_updates_during_scan(wav_path, capsys):
    """Show the user what we are doing while we scan."""
    _do_dir_scan(Audiobook(), wav_path)
    output = capsys.readouterr()
    assert (output.out == "Collecting file data...\nFile scan complete.\n")


def test_concatenation(m4a_path):
    """Concatenate all the files into one."""
    b = Audiobook()
    temp_files = natsorted(m4a_path.glob("*"))
    out_file_path = b._concatenate_files(temp_files)

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"
    assert not probe.tags.get('title')


def test_filenames_with_single_quotes(m4a_path):
    """Create a well-formed file when one of the inputs has quotes in its name."""
    orig_file = m4a_path / "2 - 220Hz.m4a"
    quote_file = m4a_path / "2 - '220Hz'.m4a"
    shutil.move(orig_file, quote_file)

    b = Audiobook()
    temp_files = natsorted(m4a_path.glob("*"))
    out_file_path = b._concatenate_files(temp_files)

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"  # If the file is broken, this will be shorter
    assert not probe.tags.get('title')


def test_add_chapter_info(m4a_path):
    """Add chapter info to a file."""
    file = m4a_path / "4 - 440Hz.m4a"
    b = Audiobook()
    b.chapters.append(SegmentData(title="Sound", start_time=0, end_time=2.499))
    b.chapters.append(SegmentData(title="Silence", start_time=2.5, end_time=4.999))
    out_file_path = b._add_chapter_info(file)

    probe = ffprobe.run_probe(out_file_path)
    assert probe
    assert len(probe.chapters) == 2
    assert probe.chapters[0]['tags']['title'] == "Sound"
    assert probe.chapters[1]['tags']['title'] == "Silence"


def test_ffmpeg_fail():
    """Handle it when an ffmpeg run fails."""
    b = Audiobook()
    tmp_path = b._tmp_path
    with testhelpers.expect_exit():
        b._run_ffmpeg(["ffmpeg", "-invalid_option"], "Make ffmpeg fail.")

    # Make sure tmp path is gone
    assert not tmp_path.exists()


def test_ffmpeg_fail_keep_temp_files():
    """Handle it when an ffmpeg run fails."""
    b = Audiobook()
    b.keep_temp_files = True
    canary = Path(b._tmp_path) / "canary"
    open(canary, 'a').close()
    with testhelpers.expect_exit():
        b._run_ffmpeg(["ffmpeg", "-invalid_option"], "Make ffmpeg fail.")
    expected_files = ["canary"]
    testhelpers.check_output_folder(b._tmp_path, expected_files, check_func=testhelpers.assert_file_path_is_file)


def test_bind_no_files(capsys):
    """Show an error when there are no files to bind."""
    b = Audiobook()
    assert (b.bind("/not-a-real-file") is False)
    output = capsys.readouterr()
    assert ("Nothing to bind" in output.out)


def test_bind_keep_temp_files(mp3_path, tmp_path, test_data_path, capsys):
    """Keep all temp files when asked."""
    out_file_path = tmp_path / "Book.m4b"
    b = Audiobook(
        cover=test_data_path / "cover.png",
        keep_temp_files=True
    )
    b.add_chapters_from_directory(mp3_path)
    b.bind(out_file_path)

    # Make sure we told the user where the temp dir is
    output = capsys.readouterr()
    assert ("Temp files can be found at" in output.out)

    # Make sure the files we expect are there
    expected_files = (
        "0_1 - 110Hz.m4a",
        "1_2 - 220Hz.m4a",
        "2_3 - 330Hz.m4a",
        "3_4 - 440Hz.m4a",
        "4_5 - 550Hz.m4a",
        "5_6 - 660Hz.m4a",
        "6_7 - 770Hz.m4a",
        "7_8 - 880Hz.m4a",
        "long.m4a",
        "filelist",
        "coverless.m4b",
        "finished.m4b",
        "ffmetadata",
    )
    testhelpers.check_output_folder(b._tmp_path, expected_files, check_func=testhelpers.assert_file_path_is_file)


def test_bind_coverless(mp3_path, tmp_path):
    """Create an audiobook without a cover."""
    out_file_path = tmp_path / "Book.m4b"
    b = Audiobook()
    b.add_chapters_from_directory(mp3_path)
    b.bind(out_file_path)

    # Verify output
    probe = ffprobe.run_probe(out_file_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "1"
    assert probe.chapters[1]['tags']['title'] == "2"
    assert probe.chapters[2]['tags']['title'] == "3"
    assert probe.chapters[3]['tags']['title'] == "4"
    assert probe.chapters[4]['tags']['title'] == "5"
    assert probe.chapters[5]['tags']['title'] == "6"
    assert probe.chapters[6]['tags']['title'] == "7"
    assert probe.chapters[7]['tags']['title'] == "8"

    # Check duration
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"

    # Check metadata
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"


def test_bind_covered(mp3_path, tmp_path, test_data_path):
    """Create an audiobook with a cover."""
    out_file_path = tmp_path / "Book.m4b"
    b = Audiobook(cover=test_data_path / "cover.png")
    b.add_chapters_from_directory(mp3_path)
    b.bind(out_file_path)

    # Verify output
    probe = ffprobe.run_probe(out_file_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "1"
    assert probe.chapters[1]['tags']['title'] == "2"
    assert probe.chapters[2]['tags']['title'] == "3"
    assert probe.chapters[3]['tags']['title'] == "4"
    assert probe.chapters[4]['tags']['title'] == "5"
    assert probe.chapters[5]['tags']['title'] == "6"
    assert probe.chapters[6]['tags']['title'] == "7"
    assert probe.chapters[7]['tags']['title'] == "8"

    # Check duration
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"

    # Check metadata
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"

    # Check cover
    assert probe.streams[2]['codec_name'] == "png"


def test_rebind_single_file(chaptered_audio_file_path, tmp_path):
    """Rebind an audiobook with changed chapter metadata."""
    b = Audiobook()
    b.add_chapters_from_chaptered_file(chaptered_audio_file_path)
    b.chapters[2].title = "NEW"
    out_path = tmp_path / "rebind.m4b"
    b.bind(out_path)

    # Verify output
    probe = ffprobe.run_probe(out_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "110Hz - Loud"
    assert probe.chapters[1]['tags']['title'] == "110Hz - Soft"
    assert probe.chapters[2]['tags']['title'] == "NEW"
    assert probe.chapters[3]['tags']['title'] == "220Hz - Soft"
    assert probe.chapters[4]['tags']['title'] == "330Hz - Loud"
    assert probe.chapters[5]['tags']['title'] == "330Hz - Soft"
    assert probe.chapters[6]['tags']['title'] == "440Hz - Loud"
    assert probe.chapters[7]['tags']['title'] == "440Hz - Soft"

    # Check duration
    assert probe.format['duration'] == "20.000000"
    assert probe.audio['duration'] == "20.000000"

    # Check metadata
    assert probe.tags['title'] == "Chaptered Audio"
    assert probe.tags['artist'] == "m4b-util"
    assert probe.tags['album'] == "Chaptered Audio"
    assert probe.tags['date'] == "2022"


def test_rebind_single_file_selections(chaptered_audio_file_path, tmp_path):
    """Create an audiobook out of selections from a single file."""
    b = Audiobook()
    b.add_chapters_from_chaptered_file(chaptered_audio_file_path)
    del b.chapters[2]
    out_path = tmp_path / "selections.m4b"
    b.bind(out_path)
    # Verify output
    probe = ffprobe.run_probe(out_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 7
    assert probe.chapters[0]['tags']['title'] == "110Hz - Loud"
    assert probe.chapters[1]['tags']['title'] == "110Hz - Soft"
    assert probe.chapters[2]['tags']['title'] == "220Hz - Soft"
    assert probe.chapters[3]['tags']['title'] == "330Hz - Loud"
    assert probe.chapters[4]['tags']['title'] == "330Hz - Soft"
    assert probe.chapters[5]['tags']['title'] == "440Hz - Loud"
    assert probe.chapters[6]['tags']['title'] == "440Hz - Soft"

    # Check duration
    assert probe.format['duration'] == "17.520333"
    assert probe.audio['duration'] == "17.520333"

    # Check metadata
    assert probe.tags['title'] == "Chaptered Audio"
    assert probe.tags['artist'] == "m4b-util"
    assert probe.tags['album'] == "Chaptered Audio"
    assert probe.tags['date'] == "2022"


def test_bad_cover(tmp_path, m4a_path):
    """Throw an error if we can't add the specified cover."""
    out_file_path = tmp_path / "Book.m4b"
    cover_path = tmp_path / "not-really-a-cover.png"
    open(cover_path, 'a').close()
    b = Audiobook(cover=cover_path)
    b.add_chapters_from_directory(m4a_path)
    with testhelpers.expect_exit():
        b.bind(out_file_path)


def test_bind_titleless_segments(tmp_path, mp3_path):
    """Allow binding, even if a segment data has no title.

    In real life, this is unlikely to happen, since all the scanner methods set the title, but it is possible to set
    the segment list manually, so we need to handle it.
    """
    out_file_path = tmp_path / "Book.m4b"
    b = Audiobook()
    b.add_chapters_from_directory(mp3_path)
    # Remove Chapter titles before we bind.
    for chapter in b.chapters:
        chapter.title = None
    b.bind(out_file_path)

    # Verify output
    probe = ffprobe.run_probe(out_file_path)
    assert probe

    # Check chapters
    assert len(probe.chapters) == 8
    assert probe.chapters[0]['tags']['title'] == "1"
    assert probe.chapters[1]['tags']['title'] == "2"
    assert probe.chapters[2]['tags']['title'] == "3"
    assert probe.chapters[3]['tags']['title'] == "4"
    assert probe.chapters[4]['tags']['title'] == "5"
    assert probe.chapters[5]['tags']['title'] == "6"
    assert probe.chapters[6]['tags']['title'] == "7"
    assert probe.chapters[7]['tags']['title'] == "8"

    # Check duration
    assert probe.format['duration'] == "40.021333"
    assert probe.audio['duration'] == "40.021333"

    # Check metadata
    assert probe.tags['title'] == "None"
    assert probe.tags['artist'] == "None"
    assert probe.tags['album'] == "None"
    assert probe.tags['date'] == "None"
    assert probe.tags['genre'] == "Audiobook"


def test_bind_non_backed_segment(tmp_path, mp3_path, capsys):
    """Alert the user if attempting to bind with a non-backed segment."""
    out_file_path = tmp_path / "Book.m4b"
    b = Audiobook()
    b.add_chapters_from_directory(mp3_path)
    # Remove backing files before we bind.
    for chapter in b.chapters:
        chapter.backing_file = None
    assert b.bind(out_file_path) is False
    output = capsys.readouterr()
    assert "Cannot bind a non-backed segment." in output.out


def test_filelist_duration_failure(mkv_file_path, capsys):
    """Warn user if we fail to find duration of a file while adding chapters."""
    b = Audiobook()
    b.add_chapters_from_filelist([mkv_file_path])
    output = capsys.readouterr()
    assert "Failed to determine duration" in output.out


def test_add_fake_chaptered_file(fake_file, capsys):
    """Warn user if we cannot read the chaptered file while adding."""
    b = Audiobook()
    b.add_chapters_from_chaptered_file(fake_file)
    output = capsys.readouterr()
    assert "Unable to parse" in output.out
    assert len(b.chapters) == 0


def test_get_cover_in_chapter_dir(test_data_path, m4a_path):
    """Search out image files named "cover" when importing chapters from a directory."""
    cover_file_path = m4a_path / "cover.png"
    shutil.copy(test_data_path / "cover.png", cover_file_path)
    _do_dir_scan(
        Audiobook(),
        m4a_path,
        cover=cover_file_path
    )


def test_noncover_image_in_chapter_dir(test_data_path, m4a_path):
    """Ignore images not named "cover" when importing chapters from a directory."""
    cover_file_path = m4a_path / "not-a-cover.png"
    shutil.copy(test_data_path / "cover.png", cover_file_path)
    _do_dir_scan(
        Audiobook(),
        m4a_path,
        cover=None
    )


def test_wrong_file_ext_cover_image_in_chapter_dir(test_data_path, m4a_path):
    """Ignore cover images with the wrong extensions when importing chapters from a directory."""
    cover_file_path = m4a_path / "cover.tiff"
    shutil.copy(test_data_path / "cover.png", cover_file_path)
    _do_dir_scan(
        Audiobook(),
        m4a_path,
        cover=None
    )


def test_tmp_path():
    """Create a temporary directory when we need it, even if it was removed."""
    b = Audiobook()
    # Create the temp dir by accessing the property
    tmp_path = b._tmp_path
    assert tmp_path.is_dir()

    # Delete the temp dir
    shutil.rmtree(tmp_path)
    assert not tmp_path.exists()

    # Recreate the deleted temp dir by accessing the property again
    new_tmp_path = b._tmp_path
    assert new_tmp_path == tmp_path
    assert tmp_path.is_dir()


def test_set_metadata():
    """Set metadata on an audiobook."""
    b = Audiobook()
    metadata = r"""
    ;FFMETADATA1
    major_brand=M4A
    minor_version=512
    compatible_brands=M4A isomiso2
    title=Super Secret: The Title
    artist=A Great Author
    album=Super Great Album
    date=2011
    genre=Audiobook
    [CHAPTER]
    TIMEBASE=1/1000
    START=0
    END=330987
    title=Opening Credits
    [CHAPTER]
    TIMEBASE=1/1000
    START=330987
    END=335943
    title=Part I
    """

    b.metadata = metadata

    assert b.title == "Super Secret: The Title"
    assert b.author == "A Great Author"
    assert b.date == "2011"

    # Check chapters
    assert len(b.chapters) == 2
    assert b.chapters[0].title == "Opening Credits"
    assert b.chapters[0].start_time == 0
    assert b.chapters[0].end_time == 330.987
    assert b.chapters[1].title == "Part I"
    assert b.chapters[1].start_time == 330.987
    assert b.chapters[1].end_time == 335.943



"""Segment Data."""
from dataclasses import dataclass
from pathlib import Path


@dataclass
class SegmentData:
    """Represents a single segment from an audio file, such as a chapter or a non-silence area.

    Variables dealing with times should be in seconds.
    """
    start_time: float
    end_time: float
    id: int = None
    title: str = None
    backing_file: Path = None
    file_start_time: float = None
    file_end_time: float = None

    def __setattr__(self, key, value):
        """Make sure all the 'time' variables are rounded to three decimal places (miliseconds)."""
        if "time" in key and value is not None:
            value = round(value, 3)
        super().__setattr__(key, value)
